theory Cooperation_Program
imports LTS Mapping Relative_Rewriting Well_Quasi_Orders Indexed_Rewriting List_Order Lexicographic_Extension
begin

subsection {* Infinite Chains *}

context pretheory begin

interpretation indexed_rewriting transition_step.

definition transitions_between ::
"('f, 'a, 't, 'b) transition set \<Rightarrow> ('a, 't, 'd, 'b) state \<Rightarrow> ('a, 't, 'd, 'b) state
\<Rightarrow> ('f, 'a, 't, 'b) transition set"
where
  "transitions_between \<tau>s s1 s2 = { \<tau> | \<beta> \<tau>. 
      assignment (valuation s1)
    \<and> assignment (valuation s2)
    \<and> assignment \<beta> \<and> \<tau> \<in> \<tau>s
    \<and> source \<tau> = location s1
    \<and> target \<tau> = location s2
    \<and> (pre_post_inter (valuation s1) (valuation s2) \<beta> \<Turnstile> transition.formula \<tau> )}"

lemma transitions_between_transition_step :
  assumes "\<tau> \<in> \<tau>s"
  shows  "(s1,s2) \<in> transition_step \<tau> \<longleftrightarrow> \<tau> \<in> transitions_between \<tau>s s1 s2"
unfolding transition_step_def transitions_between_def Let_def using assms by auto

(* not used
context fixes P::"('f,'v,'t,'l) lts" begin

definition infinite_run :: "(nat \<Rightarrow> ('v,'t,'d,'l) state) \<Rightarrow> bool" where
  "infinite_run seq \<equiv> seq 0 \<in> initial_states P \<and> chain (transition P) seq"

lemma infinite_runI[intro]:
  "seq 0 \<in> initial_states P \<Longrightarrow> chain (transition P) seq \<Longrightarrow> infinite_run seq"
  by (auto simp: infinite_run_def)

lemma infinite_runD[dest]:
  assumes "infinite_run seq"
  shows "seq 0 \<in> initial_states P" and "chain (transition P) seq"
  using assms by (auto simp: infinite_run_def)

lemma infinite_run_iff_nonTerminating :
  "(\<exists> seq . infinite_run seq) \<longleftrightarrow> \<not> lts_termination P"
unfolding SN_on_def infinite_run_def by metis
(*  :: "('f,'v,'t,'l) lts \<Rightarrow> (nat \<Rightarrow> ('v,'t,'d,'l) state) \<Rightarrow> nat \<Rightarrow> 't set" *)


lemma infinite_run_takes_transitions:
  assumes "infinite_run seq"
  shows "({} \<noteq> (transitions_between (transitions P) (seq n) (seq (Suc n))))"
proof -
  from assms have "(seq n, seq (Suc n)) \<in> transition P" by (simp add: pretheory.infinite_runD(2))
  then obtain \<tau> \<alpha>1 \<alpha>2 \<beta> where "seq n = State \<alpha>1 (source \<tau>)" "seq (Suc n) = State \<alpha>2 (target \<tau>)"
   "assignment \<alpha>1" "assignment \<alpha>2" "assignment \<beta>" "\<tau> \<in> transitions P"
   "pre_post_inter \<alpha>1 \<alpha>2 \<beta> \<Turnstile> transition.formula \<tau>"
  unfolding transition_def transition_step_def Let_def by (cases "seq n", cases "seq (Suc n)", auto)
  thus ?thesis unfolding transitions_between_def by auto
qed

lemma transitions_between_valid_transitions:
  shows "transitions_between \<tau>s (seq n) (seq (Suc n)) \<subseteq> \<tau>s"
  unfolding transitions_between_def by auto


abbreviation recurring_transition where
  "recurring_transition seq \<tau>
  \<equiv> INFM n. \<tau> \<in> transitions_between (transitions P) (seq n) (seq (Suc n))"

definition restricted_to_set where
  "restricted_to_set seq \<tau>s \<equiv> (Ball \<tau>s (recurring_transition seq)) \<and>
  (\<exists> n. \<forall> i > n. (\<exists> \<tau>. \<tau> \<in> transitions_between (transitions P) (seq n) (seq (Suc n)) \<and> \<tau> \<in> \<tau>s))"

definition cutpoint_exists where
  "cutpoint_exists seq
  \<equiv> (\<exists> \<tau>s. restricted_to_set seq \<tau>s)"

lemma every_run_has_a_cutpoint:
  assumes "infinite_run seq"
          "finite (transitions P)"
  shows "cutpoint_exists seq"
proof -
  have finPow: "finite (Pow (transitions P))" using assms by simp
  let "?stepsAfter j" = "transitions_between (transitions P) (seq (j)) (seq (Suc (j)))"
  have inPow: "\<And> n. ?stepsAfter n \<in> Pow (transitions P)"
    using assms(1) transitions_between_valid_transitions by simp
  have "\<exists>k. \<forall>i>k. INFM j. (?stepsAfter j = ?stepsAfter i)" using inf_many[OF finPow inPow].
  then obtain k where k:"\<And> i. i > k \<Longrightarrow> INFM j. (?stepsAfter j = ?stepsAfter i)" by auto
  let ?\<tau>s = "\<Union> {?stepsAfter j | j . j > k}" 
  { fix j
    assume a:"k < j"
    then obtain \<tau> where j: "\<tau> \<in> ?stepsAfter j" using infinite_run_takes_transitions assms(1) by blast
    have tauInUnion:"\<tau> \<in> \<Union>{?stepsAfter j |j. k < j}" using j a by auto
    hence "\<exists> \<tau>. \<tau> \<in> ?stepsAfter j \<and> \<tau> \<in> ?\<tau>s" using j by auto
  }
  hence c1:"\<And> n. n>k \<Longrightarrow> \<exists>\<tau>. \<tau> \<in> ?stepsAfter n \<and> \<tau> \<in> ?\<tau>s" by simp
  { fix \<tau>
    assume "\<tau> \<in> \<Union>{?stepsAfter j |j. k < j}"
    then obtain j where j: "\<tau> \<in> ?stepsAfter j \<and> k<j" by auto
    hence "\<exists>\<^sub>\<infinity>n. \<tau> \<in> ?stepsAfter n" using k j INFM_mono by fastforce
  }
  hence c2: "\<And> \<tau>. \<tau>\<in>?\<tau>s \<Longrightarrow> \<exists>\<^sub>\<infinity>n. \<tau> \<in> ?stepsAfter n" by simp
  show ?thesis using c1 c2 unfolding cutpoint_exists_def restricted_to_set_def by blast
qed

end
*)

lemma transition_subset: "subset_lts P Q \<Longrightarrow> transition P \<subseteq> transition Q"
  by (drule subset_ltsD(2), auto simp: transition_def transition_step_def Let_def)

lemma initial_states_subset: "subset_lts P Q \<Longrightarrow> initial_states P = initial_states Q"
  by (drule subset_ltsD(1), auto simp: initial_states_def)

end

subsection {* Cooperation problem *}

datatype 'l sharp = Flat (run_sharp : 'l) | Sharp (run_sharp : 'l)
derive "show" sharp
type_synonym ('f,'v,'t,'l) cooperation_problem = "('f,'v,'t,'l sharp) lts"

definition "sharp_transition \<tau> \<equiv>
  Transition (Sharp (source \<tau>)) (Sharp (target \<tau>)) (transition.formula \<tau>)"
definition "flat_transition \<tau> \<equiv>
  Transition (Flat (source \<tau>)) (Flat (target \<tau>)) (transition.formula \<tau>)"

fun is_sharp where "is_sharp (Flat _) = False" | "is_sharp (Sharp _) = True"
fun is_flat where "is_flat (Flat _) = True" | "is_flat (Sharp _) = False"

lemma is_sharp_imp_not_is_flat: "is_sharp l \<Longrightarrow> \<not> is_flat l" by (cases l, auto)

abbreviation "is_sharp_transition \<tau> \<equiv> is_sharp (source \<tau>)"
abbreviation "is_flat_transition \<tau> \<equiv> is_flat (source \<tau>)"

lemma source_sharp_transition[simp]: "source (sharp_transition \<tau>) = Sharp (source \<tau>)"
  and target_sharp_transition[simp]: "target (sharp_transition \<tau>) = Sharp (target \<tau>)"
  and formula_sharp_transition[simp]: "transition.formula (sharp_transition \<tau>) = transition.formula \<tau>"
 unfolding sharp_transition_def by auto

lemma source_flat_transition[simp]: "source (flat_transition \<tau>) = Flat (source \<tau>)"
  and target_flat_transition[simp]: "target (flat_transition \<tau>) = Flat (target \<tau>)"
  and formula_flat_transition[simp]: "transition.formula (flat_transition \<tau>) = transition.formula \<tau>"
 unfolding flat_transition_def by auto

abbreviation "sharp_state s \<equiv> relabel_state Sharp s"
abbreviation "flat_state s \<equiv> relabel_state Flat s"

fun sharp_state_pair where "sharp_state_pair (l,r) = (sharp_state l, sharp_state r)"

context pretheory begin

interpretation indexed_rewriting transition_step.

lemma
  fixes s t \<tau>
  assumes Rstep: "(s,t) \<in> transition_step \<tau>"
  shows sharp_transition: "(sharp_state s, sharp_state t) \<in> transition_step (sharp_transition \<tau>)" (is ?t0)
    and flat_transition: "(flat_state s, flat_state t) \<in> transition_step (flat_transition \<tau>)" (is ?t1)
proof-
  note * = mem_transition_stepD[OF Rstep]
  then obtain \<gamma>
    where \<gamma>: "assignment \<gamma>"
      and sat: "pre_post_inter (valuation s) (valuation t) \<gamma> \<Turnstile> transition.formula \<tau>"
    by auto
  show ?t0 apply (intro mem_transition_stepI[OF _ _ \<gamma>])
    using * sat unfolding sharp_transition_def by auto
  show ?t1 apply (intro mem_transition_stepI[OF _ _ \<gamma>])
    using * sat unfolding flat_transition_def by auto
qed

interpretation sharp_morphism: rule_morphism sharp_state _ sharp_transition transition_step transition_step
  by (standard, auto simp: sharp_transition)

interpretation flat_morphism: rule_morphism flat_state _ flat_transition transition_step transition_step
  by (standard, auto simp: flat_transition)


lemma sharp_transition_id:
  "transition_step (sharp_transition \<tau>) = sharp_state_pair ` (transition_step \<tau>)" (is "?L = ?R")
proof-
  {
    fix s t assume "(s,t) \<in> transition_step \<tau>"
    then have "sharp_state_pair (s,t) \<in> ?L"
      unfolding sharp_state_pair.simps by(intro sharp_transition, auto)
  }
  moreover
  {
    fix s t assume "(s,t) \<in> ?L"
    note * = mem_transition_stepD[OF this]
    let ?\<tau> = "sharp_transition \<tau>"
    from * have loc: "location s = source ?\<tau>" "location t = target ?\<tau>" by auto
    then obtain l r where Slr: "Sharp l = location s" "Sharp r = location t" by auto
    with loc have lr: "l = source \<tau>" "r = target \<tau>" by auto
    let ?s = "State (valuation s) l"
    let ?t = "State (valuation t) r"
    from * have ass: "state ?s" "state ?t" by auto
    from * obtain \<gamma>
    where \<gamma>: "assignment \<gamma>"
      and sat: "pre_post_inter (valuation s) (valuation t) \<gamma> \<Turnstile> transition.formula ?\<tau>"
      by auto
    from Slr have st: "(s,t) = sharp_state_pair (?s,?t)" by (simp)
    have "(State (valuation s) l, State (valuation t) r) \<in> transition_step \<tau>"
      using sat mem_transition_stepI[OF ass \<gamma>]
      unfolding st lr by auto
    hence "(s,t) \<in> ?R" unfolding st by blast
  }
  ultimately show ?thesis by auto
qed

lemma sharp_transitions:
  assumes Rstep: "(s,t) \<in> transitions_step \<tau>s"
  shows "(sharp_state s, sharp_state t) \<in> transitions_step (sharp_transition ` \<tau>s)"
proof-
  from Rstep obtain \<tau>
    where \<tau>: "\<tau> \<in> \<tau>s" and step: "(s,t) \<in> transition_step \<tau>" by force
  show ?thesis
  proof
    from \<tau> show "sharp_transition \<tau> \<in> sharp_transition ` \<tau>s" by auto
    from step show "(sharp_state s, sharp_state t) \<in> transition_step (sharp_transition \<tau>)"
      by (rule sharp_transition)
  qed
qed

definition subset_of_identity :: "('f,'v trans_var,'t) exp list list \<Rightarrow> bool" where
"subset_of_identity \<phi> \<equiv> \<forall> v. assignment v \<longrightarrow> (pre_post_inter v v v \<Turnstile> \<phi>)"

lemma subset_of_identity_transition_step:
  assumes "subset_of_identity \<phi>"
  and "assignment \<alpha>"
shows "(State \<alpha> a, State \<alpha> b) \<in> transition_step (Transition a b \<phi>)"
  using assms(1)[unfolded subset_of_identity_def] assms(2) by auto

lemma subset_of_identity_transitions:
  assumes "Transition a b \<phi> \<in> transitions P"
  and "subset_of_identity \<phi>"
  and "assignment \<alpha>"
shows "(State \<alpha> a, State \<alpha> b) \<in> transitions_step (transitions P)"
  using subset_of_identity_transition_step[OF assms(2-3), of a b] assms(1) by auto


abbreviation "sharp_transitions_of P \<equiv> { \<tau> \<in> transitions P. is_sharp_transition \<tau> }"
abbreviation "flat_transitions_of P \<equiv> { \<tau> \<in> transitions P. is_flat_transition \<tau> }"


abbreviation "cooperation_SN P \<equiv> cooperation_SN_on (flat_transitions_of P) (sharp_transitions_of P) (initial_states P)"

lemma cooperation_SN_direct:
  "cooperation_SN P \<longleftrightarrow> cooperation_SN_on {} (sharp_transitions_of P) (transitions_steps (flat_transitions_of P) `` initial_states P)"
  unfolding cooperation_SN_on_image..

context fixes P :: "('f,'v,'t,'l) cooperation_problem" begin

definition exists_cut :: "'l \<Rightarrow> bool" where
"exists_cut l \<equiv> (\<exists> \<phi>. Transition (Flat l) (Sharp l) \<phi> \<in> transitions P \<and> subset_of_identity \<phi>)"

lemma cut_transition:
  assumes "exists_cut l"
  and "assignment \<alpha>"
  shows "(State \<alpha> (Flat l), State \<alpha> (sharp.Sharp l)) \<in> transition P"
using assms unfolding exists_cut_def transition_def using subset_of_identity_transitions by auto

definition "copy_prog R \<equiv>
  lts.initial P = Flat (lts.initial R) \<and>
  (\<forall> l \<in> nodes_lts R. exists_cut l) \<and>
  (\<forall> \<tau> \<in> transitions R. flat_transition \<tau> \<in> transitions P \<and> sharp_transition \<tau> \<in> transitions P)"

context fixes R :: "('f,'v,'t,'l) lts" assumes copy: "copy_prog R" begin

private lemmas * = copy[unfolded copy_prog_def]

lemma copy_progD1: "lts.initial P = Flat (lts.initial R)" by (insert *, auto)
lemma copy_progD2: "l \<in> nodes_lts R \<Longrightarrow> exists_cut l" by (insert *, auto)
lemma copy_progD3:
  assumes "\<tau> \<in> transitions R"
  shows "flat_transition \<tau> \<in> transitions P" and "sharp_transition \<tau> \<in> transitions P"
  by (insert assms *, auto)
end

lemmas copy_progD[dest] = copy_progD1 copy_progD2 copy_progD3

lemma copy_progI[intro]:
  assumes "lts.initial P = Flat (lts.initial R)"
      and "\<And>l. l \<in> nodes_lts R \<Longrightarrow> exists_cut l"
      and "\<And>\<tau>. \<tau> \<in> transitions R \<Longrightarrow> flat_transition \<tau> \<in> transitions P"
      and "\<And>\<tau>. \<tau> \<in> transitions R \<Longrightarrow> sharp_transition \<tau> \<in> transitions P"
  shows "copy_prog R" unfolding copy_prog_def using assms by auto

lemma initial_states_copy_prog[simp]: "copy_prog R \<Longrightarrow> initial_states P = flat_state ` initial_states R"
 by (auto simp: initial_states_def image_def copy_progD1)

definition cut_chain where "cut_chain (cutpoint::nat) seq i \<equiv>
  if i \<le> cutpoint then flat_state (seq i)
  else sharp_state (seq (i-1))"

lemma cut_chain_simps[simp]:
  "i \<le> cp \<Longrightarrow> cut_chain cp seq i = flat_state (seq i)"
  "\<not> (i \<le> cp) \<Longrightarrow> cut_chain cp seq i = sharp_state (seq (i-1))"
  unfolding cut_chain_def by auto

lemma flat_state_not_sharp: "flat_state s \<notin> range sharp_state" by (cases s, auto simp:)

lemma shift_cut_chain:
  "n > cp \<Longrightarrow> shift (cut_chain cp seq) n = sharp_state \<circ> shift seq (n-1)"
  by(auto simp: cut_chain_def)

lemma sharp_chain:
  assumes chain: "chain (transitions_step \<tau>s) seq"
  shows "chain (transitions_step (sharp_transition ` \<tau>s)) (sharp_state \<circ> seq)"
proof
  let ?seq = "sharp_state \<circ> seq"
  fix i
  from chain obtain \<tau>
    where \<tau>: "\<tau> \<in> \<tau>s" and step: "(seq i, seq (Suc i)) \<in> transition_step \<tau>" by force
  from step have "(?seq i, ?seq (Suc i)) \<in> transition_step (sharp_transition \<tau>)"
    by (auto intro: sharp_transition)
  with \<tau> show "(?seq i, ?seq (Suc i)) \<in> transitions_step (sharp_transition ` \<tau>s)" by auto
qed

lemma initial_cooperation_program:
  assumes fin: "finite (transitions R)"
      and copy: "copy_prog R"
      and lts: "lts R"
      and SN: "cooperation_SN P"
  shows "lts_termination R"
proof (rule ccontr)
  assume "\<not> lts_termination R"
  then obtain seq
  where init: "seq 0 \<in> initial_states R" and chain: "chain (transition R) seq" by auto
  from chain_imp_recurring[OF fin chain[unfolded transition_def]]
  obtain n \<tau>s where \<tau>s: "\<tau>s \<subseteq> transitions R" and rec: "recurring \<tau>s (shift seq n)" by auto

  let ?\<tau>s = "sharp_transition ` \<tau>s"
  let ?seq = "cut_chain n seq"

  from SN show False
  proof (elim cooperation_SN_onE)
    show "?seq 0 \<in> initial_states P"  using init copy by auto
    show "cooperation_chain (flat_transitions_of P) (sharp_transitions_of P) ?seq"
    proof
      show "recurring ?\<tau>s (shift ?seq (Suc n))" (is "recurring _ ?s'")
        using sharp_morphism.recurring_morphism[OF rec] by(subst shift_cut_chain, auto)
    next
      fix i assume i: "i < Suc n"
      show "(?seq i, ?seq (Suc i)) \<in> transitions_step (flat_transitions_of P)"
      proof (cases "i = n")
        case True
          obtain \<alpha> l where 1: "seq n = State \<alpha> l" by (rule state.exhaust)
          have 2: "(seq n, seq (Suc n)) \<in> transition R" using chain by auto
          from mem_transition_location[OF 2] have "l \<in> nodes_lts R" by (auto simp: 1)
          moreover from 2 have "assignment \<alpha>" by (auto simp: 1)
          ultimately have "(flat_state (seq n), sharp_state (seq n)) \<in> transition P"
            unfolding 1 
            by (auto intro!: cut_transition[OF copy_progD2[OF copy]])
          then obtain \<tau>
          where \<tau>: "\<tau> \<in> transitions P"
            and step: "(flat_state (seq n), sharp_state (seq n)) \<in> transition_step \<tau>" by auto
          then have "source \<tau> = Flat l" unfolding 1 by auto
          with \<tau> step True show ?thesis by (auto intro: exI[of _ \<tau>])
      next
        case False
          with i have i: "i < n" by auto
          from chain obtain \<tau>
          where \<tau>: "\<tau> \<in> transitions R" and *: "(seq i, seq (Suc i)) \<in> transition_step \<tau>" by auto
          from \<tau> copy have "flat_transition \<tau> \<in> flat_transitions_of P" by auto
          with \<tau> flat_transition[OF *] i
          show ?thesis by (auto intro: exI[of _ "flat_transition \<tau>"])
      qed
    next
      show "sharp_transition ` \<tau>s \<subseteq> sharp_transitions_of P"
      proof
        fix \<tau>' assume "\<tau>' \<in> sharp_transition ` \<tau>s"
        then obtain \<tau> where "\<tau> \<in> \<tau>s" "\<tau>' = sharp_transition \<tau>" by auto
        with \<tau>s copy_progD(4)[OF copy]
        show "\<tau>' \<in> sharp_transitions_of P" by auto
      qed
    qed
  qed
qed

end

lemma add_new_sharp_location: assumes SN: "cooperation_SN P'"
  and flat_id: "flat_transitions_of P' = flat_transitions_of P"
  and init_id: "lts.initial P' = lts.initial P" 
  and sharp: "\<And> l1 l2 \<phi>. Transition l1 l2 \<phi> \<in> sharp_transitions_of P \<Longrightarrow>
    \<exists> \<chi> l. Transition l1 l \<chi> \<in> sharp_transitions_of P' 
      \<and> Transition l l2 \<phi> \<in> sharp_transitions_of P'
      \<and> subset_of_identity \<chi>"
  shows "cooperation_SN P" 
proof -
  from init_id have init: "initial_states P' = initial_states P" unfolding initial_states_def by auto
  show ?thesis
  proof (rule cooperation_SN_on_trancl_image[OF _ refl SN[unfolded init flat_id]])
    fix \<tau>
    assume mem: "\<tau> \<in> sharp_transitions_of P"
    obtain l1 l2 \<phi> where tau: "\<tau> = Transition l1 l2 \<phi>" by (cases \<tau>, auto)
    from sharp[OF mem[unfolded tau]] obtain \<chi> l where l1l: "Transition l1 l \<chi> \<in> sharp_transitions_of P'" 
      and ll2: "Transition l l2 \<phi> \<in> sharp_transitions_of P'"
      and id: "subset_of_identity \<chi>" by auto
    let ?t1 = "Transition l1 l \<chi>" let ?t2 = "Transition l l2 \<phi>" 
    from l1l ll2 have "{?t2, ?t1} \<subseteq> sharp_transitions_of P'" by auto
    moreover have "transition_step \<tau> \<subseteq> traverse {?t2, ?t1}"
    proof
      fix s t
      assume "(s, t) \<in> transition_step \<tau>" 
      from mem_transition_stepD[OF this, unfolded tau] obtain \<gamma> where 
        *: "state s" "state t" "location s = l1" "location t = l2" "assignment \<gamma>"
           "\<delta> (valuation s) (valuation t) \<gamma> \<Turnstile> \<phi>" by auto
      let ?u = "State (valuation s) l" 
      from l1l have "Transition l1 l \<chi> \<in> transitions P'" by auto
      from subset_of_identity_transition_step[OF id, of "valuation s" l1 l] 
      have 1: "(s, ?u) \<in> transition_step ?t1" using * by (cases s, auto)
      have 2: "(?u, t) \<in> transition_step ?t2"
        by (rule mem_transition_stepI[OF _ *(2,5)], insert *, auto)
      show "(s,t) \<in> traverse {?t2, ?t1}" 
        by (intro mem_traverseI[of "\<lambda> i. if i = 0 then s else if i = 1 then ?u else t" _ "Suc (Suc 0)"], 
        force+, intro traversed.intros, insert 1 2, auto)
    qed
    ultimately show 
      "\<exists>ps. ps \<noteq> {} \<and> ps \<subseteq> {\<tau> \<in> transitions P'. is_sharp_transition \<tau>} \<and> transition_step \<tau> \<subseteq> traverse ps" 
      by (intro exI[of _ "{?t2, ?t1}"], auto)
  qed
qed
end


type_synonym ('f,'v,'t,'l,'tr,'err,'ok) checker
   = "('l sharp \<Rightarrow> ('f, 'v, 't) formula list) \<Rightarrow> 'tr \<times> ('f, 'v, 't, 'l sharp) transition \<Rightarrow> 'err + 'ok"

datatype ('f,'v,'t,'d,'l,'tr,'x,'err,'ok) condition_checker
  = Condition_Checker (inv_checker:"('f,'v,'t,'l,'tr,'err,'ok) checker")
                      (str_checker:"('f,'v,'t,'l,'tr,'err,'ok) checker")
                      (bnd_checker:"('f,'v,'t,'l,'tr,'err,'ok) checker")
                      (ranking_f:"('v, 't, 'd, 'l sharp) state \<Rightarrow> 'x")
                      (strict_o:"('x \<times> 'x) set")
                      (nonstrict_o:"('x \<times> 'x) set")
                      (bound:"'x")

context pretheory begin

interpretation indexed_rewriting transition_step.

definition "condition_checker inv_c str_c bnd_c rf S NS bnd \<equiv>
  (non_inf_order S NS) \<and>
  (\<forall> tr \<tau> I.
    (\<forall> l. list_all formula (I l)) \<longrightarrow>
    is_sharp_transition \<tau> \<longrightarrow>
    ( formula (transition.formula \<tau>) \<longrightarrow>
      isOK(inv_c I (tr,\<tau>)) \<longrightarrow>
      (\<forall> (s,t) \<in> transition_step \<tau>. \<forall> \<phi> \<in> set (I (location s)).
        (valuation s \<Turnstile> \<phi>) \<longrightarrow> (rf s, rf t) \<in> NS \<union> S)) \<and>
    ( formula (transition.formula \<tau>) \<longrightarrow>
      isOK(bnd_c I (tr,\<tau>)) \<longrightarrow>
      isOK(str_c I (tr,\<tau>)) \<longrightarrow>
      (\<forall> (s,t) \<in> transition_step \<tau>. \<forall> \<phi> \<in> set (I (location s)).
         (valuation s \<Turnstile> \<phi>) \<longrightarrow> (rf s, rf t) \<in> S)) \<and>
    ( formula (transition.formula \<tau>) \<longrightarrow>
      isOK(bnd_c I (tr,\<tau>)) \<longrightarrow>
      (\<forall> (s,t) \<in> transition_step \<tau>. \<forall> \<phi> \<in> set (I (location s)).
         (valuation s \<Turnstile> \<phi>) \<longrightarrow> (rf s, bnd) \<in> S)))"

lemma condition_checkerD:
  assumes "condition_checker inv_c str_c bnd_c  rf S NS bnd"
  shows "non_inf_order S NS"
    and "\<And> tr \<tau> I. (\<forall> l. list_all formula (I l)) \<Longrightarrow>
          formula (transition.formula \<tau>) \<Longrightarrow>
          is_sharp_transition \<tau> \<Longrightarrow>
          ( isOK(inv_c I (tr,\<tau>)) \<Longrightarrow>
            (\<And> s t \<phi>.
              (s,t) \<in> transition_step \<tau> \<Longrightarrow> \<phi> \<in> set (I (location s)) \<Longrightarrow>
              valuation s \<Turnstile> \<phi> \<Longrightarrow> (rf s, rf t) \<in> NS \<union> S))"
    and "\<And> tr \<tau> I. (\<forall> l. list_all formula (I l)) \<Longrightarrow>
          formula (transition.formula \<tau>) \<Longrightarrow>
          is_sharp_transition \<tau> \<Longrightarrow>
          ( isOK(bnd_c I (tr,\<tau>)) \<Longrightarrow>
            isOK(str_c I (tr,\<tau>)) \<Longrightarrow>
            (\<And> s t \<phi>.
              (s,t) \<in> transition_step \<tau> \<Longrightarrow> \<phi> \<in> set (I (location s)) \<Longrightarrow>
              valuation s \<Turnstile> \<phi> \<Longrightarrow> (rf s, rf t) \<in> S))"
    and "\<And> tr \<tau> I. (\<forall> l. list_all formula (I l)) \<Longrightarrow>
          formula (transition.formula \<tau>) \<Longrightarrow>
          is_sharp_transition \<tau> \<Longrightarrow>
          ( isOK(bnd_c I (tr,\<tau>)) \<Longrightarrow>
            (\<And> s t \<phi>.
              (s,t) \<in> transition_step \<tau> \<Longrightarrow> \<phi> \<in> set (I (location s)) \<Longrightarrow>
              valuation s \<Turnstile> \<phi> \<Longrightarrow> (rf s, bnd) \<in> S))"
  using assms[unfolded condition_checker_def] by blast+

lemma condition_checkerI[intro]:
  assumes "non_inf_order S NS"
    and "\<And> tr \<tau> I. (\<forall> l. list_all formula (I l)) \<Longrightarrow>
          formula (transition.formula \<tau>) \<Longrightarrow>
          is_sharp_transition \<tau> \<Longrightarrow>
          ( isOK(inv_c I (tr,\<tau>)) \<Longrightarrow>
            (\<And> s t \<phi>.
              (s,t) \<in> transition_step \<tau> \<Longrightarrow> \<phi> \<in> set (I (location s)) \<Longrightarrow>
              valuation s \<Turnstile> \<phi> \<Longrightarrow> (rf s, rf t) \<in> NS \<union> S))"
    and "\<And> tr \<tau> I. (\<forall> l. list_all formula (I l)) \<Longrightarrow>
          formula (transition.formula \<tau>) \<Longrightarrow>
          is_sharp_transition \<tau> \<Longrightarrow>
          ( isOK(bnd_c I (tr,\<tau>)) \<Longrightarrow>
            isOK(str_c I (tr,\<tau>)) \<Longrightarrow>
            (\<And> s t \<phi>.
              (s,t) \<in> transition_step \<tau> \<Longrightarrow> \<phi> \<in> set (I (location s)) \<Longrightarrow>
              valuation s \<Turnstile> \<phi> \<Longrightarrow> (rf s, rf t) \<in> S))"
    and "\<And> tr \<tau> I. (\<forall> l. list_all formula (I l)) \<Longrightarrow>
          formula (transition.formula \<tau>) \<Longrightarrow>
          is_sharp_transition \<tau> \<Longrightarrow>
          ( isOK(bnd_c I (tr,\<tau>)) \<Longrightarrow>
            (\<And> s t \<phi>.
              (s,t) \<in> transition_step \<tau> \<Longrightarrow> \<phi> \<in> set (I (location s)) \<Longrightarrow>
              valuation s \<Turnstile> \<phi> \<Longrightarrow> (rf s, bnd) \<in> S))"
  shows "condition_checker inv_c str_c bnd_c rf S NS bnd"
  using assms unfolding condition_checker_def by(auto,blast)


(* 
  and is_ok: "isOK (delete_condition rf bnd hint I (tr, b))"
  and formula_ok: "list_all LIA_checker.formula (I l)"
  and isFormula: "LIA_checker.formula (transition.formula b)" *)

lemma trivial_cooperation_program:
  assumes emp: "sharp_transitions_of P = {}" shows "cooperation_SN P" unfolding emp by auto

lemma ranking_function_via_non_inf_order_main: fixes P :: "('f,'v,'t,'l sharp) lts"
  assumes ni: "non_inf_order S NS"
  and TS: "\<forall>\<tau> \<in> TS. is_sharp_transition \<tau>"
  and TB: "\<forall>\<tau> \<in> TB. is_sharp_transition \<tau>"
  and NS: "\<And> s t \<tau>. \<tau> \<in> sharp_transitions_of P \<Longrightarrow> s \<in> reachable_states P
             \<Longrightarrow> (s,t) \<in> transition_step \<tau> \<Longrightarrow> (f s, f t) \<in> NS \<union> S"
  and S: "\<And> s t \<tau>. \<tau> \<in> TS \<Longrightarrow> s \<in> reachable_states P
             \<Longrightarrow> (s,t) \<in> transition_step \<tau> \<Longrightarrow> (f s, f t) \<in> S"
  and B: "\<And> s t \<tau>. \<tau> \<in> TB \<Longrightarrow> s \<in> reachable_states P
             \<Longrightarrow> (s,t) \<in> transition_step \<tau> \<Longrightarrow> (f s, bnd) \<in> NS \<union> S"
  and SN: "cooperation_SN (delete_transitions P TS)"
          "cooperation_SN (delete_transitions P TB)"
  shows "cooperation_SN P"
  unfolding cooperation_SN_on_as_not_recurrent_on
proof (intro allI impI notI, elim recurrent_onE)

  interpret non_inf_order S NS by fact

  fix seq \<tau>s
  assume \<tau>s: "\<tau>s \<subseteq> sharp_transitions_of P"
     and init: "seq 0 \<in> transitions_steps (flat_transitions_of P) `` initial_states P"
     and tang: "recurring \<tau>s seq"

  from recurring_imp_chain[OF tang]
  have chain: "\<forall> i. \<exists> \<tau>. \<tau> \<in> \<tau>s \<and> step seq i \<in> transition_step \<tau>" by auto
  from recurring_imp_INFM[OF tang]
  have rec: "\<forall>\<tau> \<in> \<tau>s. \<exists>\<^sub>\<infinity>i. step seq i \<in> transition_step \<tau>" by auto

  from choice[OF chain] obtain \<tau> where \<tau>ts: "\<And> i. \<tau> i \<in> \<tau>s"
    and step: "\<And> i. step seq i \<in> transition_step (\<tau> i)" by blast
  with \<tau>s have \<tau>: "\<And> i. \<tau> i \<in> sharp_transitions_of P" by auto

  have Pchain: "step seq i \<in> transition P" for i
  proof -
    note step[of i]
    also have "transition_step (\<tau> i) \<subseteq> transitions_step (sharp_transitions_of P)"
      using \<tau>[of i] by auto
    also have "... \<subseteq> transition P" by auto
    finally show ?thesis.
  qed
  have reach: "seq i \<in> reachable_states P" for i
  proof (induct i)
    case 0
      show ?case by(rule rev_subsetD[OF init], rule Image_subset[OF rtrancl_mono], auto simp: transition_def)
  next
    case (Suc i) with Pchain[of i] show ?case by (auto intro: rtrancl_image_advance)
  qed
  def fS \<equiv> "{i. (f (seq i), f (seq (Suc i))) \<in> S}"
  def fB \<equiv> "{i. (f (seq i), bnd) \<in> NS \<union> S}"
  from non_inf_order.chain_split[of S NS fS f seq fB bnd,OF ni _ _ NS[OF \<tau> reach step]]
  have disj: "(\<forall>\<^sub>\<infinity>j. j \<notin> fS) \<or> (\<forall>\<^sub>\<infinity>j. j \<notin> fB)" unfolding fS_def fB_def by auto
  {
    fix fD TD m
    let ?P = "delete_transitions P TD"
    assume infm: "(\<forall>\<^sub>\<infinity>j. j \<notin> fD)"
      and TD: "\<forall>\<tau> \<in> TD. is_sharp_transition \<tau>"
      and SN: "cooperation_SN ?P"
      and step: "\<And> i \<tau>. \<tau> \<in> TD \<Longrightarrow> step seq i \<in> transition_step \<tau> \<Longrightarrow> i \<in> fD"
    note SN = SN 
    from TD have "\<forall> \<tau> \<in> TD. \<not> is_flat_transition \<tau>" by (auto simp: is_sharp_imp_not_is_flat)
    then have "flat_transitions_of ?P = flat_transitions_of P" by (auto simp: delete_transitions_def)
    with init have init': "seq 0 \<in> transitions_steps (flat_transitions_of ?P) `` initial_states ?P"
      by (auto simp: delete_transitions_def initial_states_def)

    from tang have "cooperation_chain {} (sharp_transitions_of ?P) seq"
      unfolding cooperation_chain_empty
    proof(intro exI conjI)
      show \<tau>s': "\<tau>s \<subseteq> sharp_transitions_of ?P"
      proof
        fix \<tau> assume \<tau>ts: "\<tau> \<in> \<tau>s"
        have "\<tau> \<notin> TD"
        proof
          assume \<tau>: "\<tau> \<in> TD"
          from infm[unfolded MOST_nat_le] obtain j where j: "\<And> k. k \<ge> j \<Longrightarrow> k \<notin> fD" by auto
          from rec[unfolded INFM_nat_le, rule_format, OF \<tau>ts] obtain k
            where "k \<ge> j" "step seq k \<in> transition_step \<tau>" by auto
          then have "k \<ge> j" and "step seq k \<in> transition_step \<tau>" by auto
          with step[OF \<tau>] reach j show False by auto
        qed
        with \<tau>ts \<tau>s show "\<tau> \<in> sharp_transitions_of ?P"
          by (auto simp: delete_transitions_def)
      qed
    qed
    with cooperation_SN_onE[OF SN(1)[unfolded cooperation_SN_direct], of seq, OF init'] have False by auto
  }
  note main = this

  from disj show False
  proof
    assume "\<forall>\<^sub>\<infinity>j. j \<notin> fS"
    from main[OF this TS SN(1)] S[OF _ reach] show False by(auto simp: fS_def)
  next
    assume "\<forall>\<^sub>\<infinity>j. j \<notin> fB"
    from main[OF this TB SN(2)] B[OF _ reach] show False by(auto simp: fB_def)
  qed
qed

lemma ranking_function_via_non_inf_order: fixes P :: "('f,'v,'t,'l sharp) lts"
  assumes ni: "non_inf_order S NS"
  and TD: "\<forall>\<tau> \<in> TD. is_sharp_transition \<tau>"
  and NS: "\<And> s t \<tau>. \<tau> \<in> sharp_transitions_of P \<Longrightarrow> s \<in> reachable_states P
    \<Longrightarrow> (s,t) \<in> transition_step \<tau> \<Longrightarrow> (f s, f t) \<in> NS \<union> S"
  and S: "\<And> s t \<tau>. \<tau> \<in> TD \<Longrightarrow> s \<in> reachable_states P
    \<Longrightarrow> (s,t) \<in> transition_step \<tau> \<Longrightarrow> (f s, f t) \<in> S \<and> (f s, bnd) \<in> NS \<union> S"
  and SN: "cooperation_SN (delete_transitions P TD)"
  shows "cooperation_SN P"
  apply (rule ranking_function_via_non_inf_order_main[OF ni TD TD NS S[THEN conjunct1] S[THEN conjunct2]])
  using SN by auto

lemma ranking_function_via_disj_invariant:
  fixes P :: "('f,'v,'t,'l sharp) lts" and S :: "'a rel"
    and rf :: "'l sharp \<Rightarrow> ('v,'t,'d) valuation \<Rightarrow> 'a"
  assumes I: "disj_invariant P I"
      and ni: "non_inf_order S NS"
      and TD: "\<forall>\<tau> \<in> TD. is_sharp_transition \<tau>"
      and NS: "\<forall> \<tau> \<in> sharp_transitions_of P. \<forall> (s, t) \<in> transition_step \<tau>. \<forall> \<phi> \<in> set (I (location s)).
        (valuation s \<Turnstile> \<phi>) \<longrightarrow> (f s, f t) \<in> NS \<union> S"
      and  S: "\<forall> \<tau> \<in> TD. \<forall> (s, t) \<in> transition_step \<tau>. \<forall> \<phi> \<in> set (I (location s)).
        (valuation s \<Turnstile> \<phi>) \<longrightarrow> (f s, f t) \<in> S \<and> (f s, bnd) \<in> NS \<union> S"
      and SN: "cooperation_SN (delete_transitions P TD)"
  shows "cooperation_SN P"
proof (rule ranking_function_via_non_inf_order[OF ni TD _ _ SN, of f])
  fix s t \<tau>
  assume reach: "s \<in> reachable_states P" and st: "(s,t) \<in> transition_step \<tau>"
  from reach disj_invariantD[OF I, of "valuation s" "location s"] obtain \<phi>
  where inv: "\<phi> \<in> set (I (location s))" "valuation s \<Turnstile> \<phi>" by auto
  { assume "\<tau> \<in> sharp_transitions_of P"
    from NS[rule_format, OF this st] inv show "(f s, f t) \<in> NS \<union> S" by auto
  }
  { assume "\<tau> \<in> TD"
    from S[rule_format,OF this st] inv show "(f s, f t) \<in> S \<and> (f s, bnd) \<in> NS \<union> S" by auto
  }
qed

lemma finite_transitions_inst[simp]: "finite (transitions (lts_of P))"
proof -
  have "finite (snd ` set (transitions_impl P))" by simp
  thus ?thesis unfolding "lts_of.simps" by simp
qed

(* Note: should a preserved property be required, you can add it here *)
definition preserved_lookup where
"preserved_lookup lookup P \<equiv> lts_impl P \<and> (\<forall>tr b. (tr, b) \<in> set (transitions_impl P) \<longrightarrow> b = lookup tr)"
definition preserved_invariant where
"preserved_invariant I P \<equiv> disj_invariant (lts_of P) I \<and> (\<forall> l. list_all formula (I l))"

definition "disj_invariant_create f \<equiv> 
  \<forall> P lookup. preserved_lookup lookup P \<longrightarrow>
  (case f lookup P of Inl _ \<Rightarrow> True
  | Inr (P2,lookup2,I) \<Rightarrow> (cooperation_SN (lts_of P2) \<longrightarrow> cooperation_SN (lts_of P))
               \<and> preserved_lookup lookup2 P2
               \<and> preserved_invariant I P2)"

definition "disj_invariant_update f \<equiv> 
  \<forall> P lookup I. (preserved_lookup lookup P \<and> preserved_invariant I P) \<longrightarrow>
  (case f I lookup P of Inl _ \<Rightarrow> True
  | Inr (P2,lookup2,I2) \<Rightarrow> (cooperation_SN (lts_of P2) \<longrightarrow> cooperation_SN (lts_of P))
               \<and> preserved_lookup lookup2 P2
               \<and> preserved_invariant I2 P2)"

lemma disj_invariant_update_SN_preserving[simp]:
  assumes "disj_invariant_create create"
  and "disj_invariant_update update"
  shows "disj_invariant_create (\<lambda>lookup P. create lookup P \<bind> (\<lambda>(P2, lookup2, I). update I lookup2 P2))"
using assms
by (auto simp: disj_invariant_create_def disj_invariant_update_def Error_Monad.bind_def
               sum.splits prod.split)

definition "disj_invariant_update_P_only f \<equiv> 
   \<forall> P lookup I. (preserved_lookup lookup P \<and> preserved_invariant I P) \<longrightarrow>
  (case f I lookup P of Inl _ \<Rightarrow> True
  | Inr P2 \<Rightarrow> (cooperation_SN (lts_of P2) \<longrightarrow> cooperation_SN (lts_of P))
           \<and> preserved_lookup lookup P2
           \<and> preserved_invariant I P2)"

definition "no_modification_create f lookup P = (case f P of Inl v \<Rightarrow> Inl v
                                              | Inr I \<Rightarrow> Inr (P,lookup,I))"

lemma no_modification_is_disj_invariant_create[simp]:
  "disj_invariant_checker f \<Longrightarrow> disj_invariant_create (no_modification_create f)"
by (auto simp:disj_invariant_create_def preserved_lookup_def lts_impl_def
              disj_invariant_checker_def no_modification_create_def preserved_invariant_def
         split:sum.splits)

definition "no_modification_update f I lookup P = (case f I lookup P of Inl v \<Rightarrow> Inl v
           | Inr P2 \<Rightarrow> Inr (P2,lookup,I))"

lemma no_modification_is_disj_invariant_update[simp]:
  "disj_invariant_update_P_only f \<Longrightarrow> disj_invariant_update (no_modification_update f)"
by (auto simp:disj_invariant_update_P_only_def disj_invariant_update_def no_modification_update_def
         split:sum.splits)

definition "no_invariants f lookup P = (case f lookup P of Inl v \<Rightarrow> Inl v
                                         | Inr (P2,lookup2,I) \<Rightarrow> Inr (lookup2,P2))"

definition cooperation_SN_preserving where
"cooperation_SN_preserving lookup f \<equiv>
 (\<forall> P. preserved_lookup lookup P \<longrightarrow>
    (case f lookup P of Inl _ \<Rightarrow> True
     | Inr (lookup2,P2) \<Rightarrow> preserved_lookup lookup2 P2 \<and>
        (cooperation_SN (lts_of P2) \<longrightarrow> cooperation_SN (lts_of P))))"

lemma no_invariants_is_cooperation_SN_preserving[simp]:
  "disj_invariant_create f \<Longrightarrow> cooperation_SN_preserving lookup (no_invariants f)"
by (auto simp: disj_invariant_create_def cooperation_SN_preserving_def no_invariants_def
         split: sum.splits)

lemma cooperation_SN_preserving:
assumes "cooperation_SN_preserving lookup a"
        "a lookup P = Inr (lookup2,P2)"
        "preserved_lookup lookup P"
shows "cooperation_SN (lts_of P2) \<Longrightarrow> cooperation_SN (lts_of P)"
      "preserved_lookup lookup2 P2"
using assms by (auto simp:cooperation_SN_preserving_def sum.splits)


fun check_termination where
"check_termination Nil lookup (P::('f,'v,'t,'l sharp,'tr)lts_impl) =
 check (list_all (\<lambda> (tr,\<tau>). \<not> is_sharp_transition \<tau>) (transitions_impl P)) (shows ''After applying the final hint, there are still sharp transitions left'')" |
"check_termination (Cons h hs) lookup P
  = do{(lookup2,P2) <- h lookup P;check_termination hs lookup2 P2}"

lemma check_termination_isOk:
  assumes "isOK (check_termination h lookup P)"
      and "\<And> lookup. list_all (cooperation_SN_preserving lookup) h"
      and "preserved_lookup lookup P"
  shows "cooperation_SN (lts_of P)"
using assms proof(induction h lookup P rule:"check_termination.induct")
  case (1 l P)
    hence "list_all (\<lambda> (tr,\<tau>). \<not> is_sharp_transition \<tau>) (transitions_impl P)" by simp
    hence *: "sharp_transitions_of (lts_of P) = {}"
      unfolding empty_Collect_eq Ball_set[symmetric] by auto
    show "cooperation_SN (lts_of P)" unfolding * by auto next
  case (2 a as lookup P)
    obtain lookup2 P2 where
      P2:"a lookup P = Inr (lookup2,P2)" "isOK (check_termination as lookup2 P2)" using 2 by auto
    hence c2:"cooperation_SN (lts_of P2) \<Longrightarrow> cooperation_SN (lts_of P)"
             "preserved_lookup lookup2 P2"
      using cooperation_SN_preserving[of lookup a P lookup2 P2] 2 P2 by auto
    show ?case using c2(1)[OF 2(1)[OF P2(1) refl P2(2) _ c2(2)]] 2(3) by simp
qed

definition check_termination_and_transition where
"check_termination_and_transition lst lookup P
  = (check (list_all (\<lambda> (tr,\<tau>). lookup tr = \<tau>) (transitions_impl P)) (shows ''Transition names should be unique'')  \<then>
     check (lts_impl P) (shows ''Non-formulas in cooperation-graph (maybe type-errors in formula?)'') \<then>
    check_termination lst lookup P)"

lemma check_termination_and_transition:
  assumes "isOK (check_termination_and_transition h lookup P)"
      and "\<And> lookup. list_all (cooperation_SN_preserving lookup) h"
  shows "cooperation_SN (lts_of P)"
proof (rule check_termination_isOk)
  note * = assms(1)[unfolded check_termination_and_transition_def]
  show "isOK (check_termination h lookup P)" using * by auto
  show "preserved_lookup lookup P" using * by (auto simp:pred_list_def preserved_lookup_def)
qed (insert assms,auto)

fun with_invariants_reduction where
"with_invariants_reduction Nil base = base" |
"with_invariants_reduction (Cons f fs) base
  = (\<lambda> lookup P. do{(P2,lookup2,I) <- with_invariants_reduction fs base lookup P
                   ;(P3,lookup3,I2) <- f I lookup2 P2
                   ;Inr (P3,lookup3,I2)})"

lemma with_invariants_reduction_is_create[simp]:
  assumes "disj_invariant_create base"
      and "list_all disj_invariant_update lst"
  shows "disj_invariant_create (with_invariants_reduction lst base)"
using assms by(induction lst,auto)

definition cooperation_cond  where
"cooperation_cond hints coopProg original
  = (check (lts_impl original) (shows ''Non-formulas in original program (maybe type-errors in formula?)'') \<then>
     check (copy_prog (lts_of coopProg) (lts_of original)) (shows ''Cooperation-graph is not a valid copy of the program'') \<then>
     check_termination_and_transition hints (transition_of coopProg) coopProg
    ) <+? (\<lambda> s. shows ''Cooperation_checker failed'')"

lemma checker_is_ok:
  assumes isok:"isOK (cooperation_cond h P R)"
      and isVaildCheck:"\<And> tr. list_all (cooperation_SN_preserving tr) h"
  shows "lts_termination (lts_of R)"
proof -
  have copy:"copy_prog (lts_of P) (lts_of R)"
    and terminate:"isOK (check_termination_and_transition h (transition_of P) P)"
    and ltsR:"lts_impl R"
    using isok unfolding cooperation_cond_def by simp_all
  show ?thesis
    using initial_cooperation_program[OF finite_transitions_inst copy lts_impl[OF ltsR]
          check_termination_and_transition[OF terminate isVaildCheck]].
qed

fun update_transitions_impl where
"update_transitions_impl (Lts_Impl i l) ts = (Lts_Impl i ts)"

lemma transitions_of_update_transitions_impl[simp]:
  "transitions_impl (update_transitions_impl P v) = v" by(cases P;simp)
lemma initials_of_update_transitions_impl[simp]:
  "initial (update_transitions_impl P v) = initial P" by(cases P;simp)

definition diff_by_label :: "('a \<times> 'b) list \<Rightarrow> 'a set \<Rightarrow> ('a \<times> 'b) list" where
"diff_by_label x y = List.filter (\<lambda> v. fst v \<notin> y) x"

lemma diff_by_label_subset [intro]:
  "set (diff_by_label a b) \<subseteq> set a"
unfolding diff_by_label_def by auto

lemma diff_by_label_preseves_distinct [simp]:
  "distinct (map fst a) \<Longrightarrow> distinct (map fst (diff_by_label a b))"
  by(induct a;auto simp:diff_by_label_def)

definition del_transitions_impl ::
    "('f,'v,'t,'l,'tr) lts_impl \<Rightarrow> 'tr list \<Rightarrow> ('f,'v,'t,'l,'tr) lts_impl" where
"del_transitions_impl P TD
 = (update_transitions_impl P (diff_by_label (transitions_impl P) (set TD)))"

lemma transitions_of_del_transitions_impl[simp]:
  "transitions_impl (del_transitions_impl P D) = diff_by_label (transitions_impl P) (set D)"
unfolding del_transitions_impl_def by simp

lemma del_transitions_impl_set[simp]:
assumes "set transs \<subseteq> set transs2"
shows "set (diff_by_label transs (fst ` set TD)) \<subseteq> set transs2"
using diff_by_label_subset[of "transs" "(fst ` set TD)"] assms by simp

lemma del_transitions_subset[simp]:
shows "set (transitions_impl (del_transitions_impl P TD)) \<subseteq> set (transitions_impl P)"
    by (simp add: pretheory.diff_by_label_subset)

lemma del_transitions_subsets:
  shows "transitions (lts_of (del_transitions_impl P TD)) \<subseteq> transitions (lts_of P)"
    and "initial_states (lts_of (del_transitions_impl P TD)) \<subseteq> initial_states (lts_of P)"
by(auto simp:del_transitions_impl_def diff_by_label_def initial_states_def)

lemma preserved_is_preserved_over_del_transitions:
  assumes "preserved_lookup lookup P"
    shows "preserved_lookup lookup (del_transitions_impl P TD)"
proof -
  have "lts_impl P \<and>
   (\<forall>tr. tr \<in> fst ` set (transitions_impl P) \<longrightarrow> (\<forall> b. (tr, b) \<in> set (transitions_impl P) \<longrightarrow> b = lookup tr))"
    using assms unfolding preserved_lookup_def by simp
  hence "lts_impl (del_transitions_impl P TD)" using pretheory.del_transitions_subset
    unfolding lts_impl_def pretheory.lts_def by fastforce
  thus ?thesis using del_transitions_subset assms unfolding preserved_lookup_def by blast
qed

lemma reachable_states_declines_by_del_transitions:
  assumes "s \<in> reachable_states (lts_of (del_transitions_impl P TD))"
  shows "s \<in> reachable_states (lts_of P)"
using assms in_mono[OF reachable_states_declines_by_subset[OF del_transitions_subsets]] by blast

lemma disj_invariant_is_preserved_over_del_transitions:
  assumes "disj_invariant (lts_of P) I"
  shows "disj_invariant (lts_of (del_transitions_impl P TD)) I"
using reachable_states_declines_by_del_transitions assms
unfolding disj_invariant_def by blast

lemma preserved_invariant_is_preserved_over_del_transitions:
  assumes "preserved_invariant I P"
    shows "preserved_invariant I (del_transitions_impl P TD)"
using assms disj_invariant_is_preserved_over_del_transitions
unfolding preserved_invariant_def by auto

lemma del_transitions_impl:
assumes inj:"(\<forall>tr b. tr \<in> fst ` set (transitions_impl P) \<longrightarrow>
           (tr, b) \<in> set (transitions_impl P) \<longrightarrow> b = lookup tr)"
    and was_SN:"cooperation_SN (lts_of (del_transitions_impl P TD))" (is "cooperation_SN ?P'")
  shows "cooperation_SN (delete_transitions (lts_of P) (lookup ` set TD))" (is "cooperation_SN ?P")
  proof(cases "P")
  case (Lts_Impl x1 x2)
  have c1:"lts_impl.initial (del_transitions_impl (Lts_Impl x1 x2) TD) = x1"
    unfolding del_transitions_impl_def by simp
  have c2:"snd ` set x2 - lookup ` set TD \<subseteq> snd ` set (diff_by_label x2 (set TD))"
    unfolding diff_by_label_def using inj[unfolded lts_impl.sel(2) Lts_Impl]
    using DiffE image_iff mem_Collect_eq prod.collapse set_filter subsetI
    by fastforce
  have "subset_lts ?P ?P'" unfolding Lts_Impl delete_transitions_def subset_lts_def using c1 c2 by simp
  then have subs: "transitions ?P \<subseteq> transitions ?P'"
        and init: "lts.initial ?P = lts.initial ?P'" unfolding subset_lts_def by auto
  from subs
  have subs: "flat_transitions_of ?P \<subseteq> flat_transitions_of ?P'" "sharp_transitions_of ?P \<subseteq> sharp_transitions_of ?P'"
    by auto
  from init have init: "initial_states ?P = initial_states ?P'" by (auto simp: initial_states_def)
  show ?thesis unfolding init by(rule cooperation_SN_on_subset[OF subs was_SN])
qed

definition eliminate_edges ::
  "string
  \<Rightarrow> (('f,'v,'t,'l sharp) disjInv \<Rightarrow> 'tr \<times> ('f,'v,'t,'l sharp) transition \<Rightarrow> shows check)
  \<Rightarrow> (('f,'v,'t,'l sharp) disjInv \<Rightarrow> 'tr \<times> ('f,'v,'t,'l sharp) transition \<Rightarrow> shows check)
  \<Rightarrow> (('f,'v,'t,'l sharp) disjInv \<Rightarrow> 'tr \<times> ('f,'v,'t,'l sharp) transition \<Rightarrow> shows check)
  \<Rightarrow> 'tr list
  \<Rightarrow> (('f,'v,'t,'l sharp) disjInv)
  \<Rightarrow> ('tr \<Rightarrow> ('f,'v,'t,'l sharp) transition)
  \<Rightarrow> ('f,'v,'t,'l sharp,'tr) lts_impl
  \<Rightarrow> shows + (('f,'v,'t,'l sharp,'tr) lts_impl)"
where
"eliminate_edges step_id invariant_condition str_c bnd_c TD I lookup P
 = do {check_allm (invariant_condition I) (filter (is_sharp_transition \<circ> snd) (transitions_impl P))
      ;check_allm (str_c I) (map (\<lambda> v. (v,lookup v)) TD)
      ;check_allm (bnd_c I) (map (\<lambda> v. (v,lookup v)) TD)
      ;Inr (del_transitions_impl P (filter (is_sharp_transition \<circ> lookup) TD))
      } <+? (\<lambda> s. shows ''Failed to eliminate edges in step: '' \<circ> shows step_id \<circ> s)"

lemma SN_order_pair_imp_non_inf_order:
  assumes "SN_order_pair S NS"
  shows "non_inf_order S NS"
  using assms[unfolded SN_order_pair_def order_pair_def
                       compat_pair_def pre_order_pair_def SN_ars_def]
  by(unfold_locales,auto simp:SN_imp_non_inf)

lemma filter_formula:
  assumes "preserved_lookup lookup P"
  shows "del_transitions_impl P (filter (is_sharp_transition \<circ> lookup) TD) = del_transitions_impl P [v\<leftarrow>TD . is_sharp_transition (lookup v) \<and> local.formula (transition.formula (lookup v))]"
proof -
  have r:"\<And> \<tau>. \<tau> \<in> set (transitions_impl P) \<Longrightarrow> local.formula (transition.formula (lookup (fst \<tau>)))"
   using assms[unfolded lts_impl_def lts_def preserved_lookup_def] by auto
  hence "\<And> x. x \<in> set (transitions_impl P) \<Longrightarrow>
         (fst x \<in> set TD \<longrightarrow> \<not> is_sharp_transition (lookup (fst x))) = (is_sharp_transition (lookup (fst x)) \<longrightarrow> fst x \<in> set TD \<longrightarrow> \<not> local.formula (transition.formula (lookup (fst x))))"
    by auto
  from filter_cong[OF refl[of "transitions_impl P"] this, of id]
  have "[v\<leftarrow>(transitions_impl P) . fst v \<in> set TD \<longrightarrow> \<not> is_sharp_transition (lookup (fst v))] =
         [v\<leftarrow>(transitions_impl P) . is_sharp_transition (lookup (fst v)) \<longrightarrow> fst v \<in> set TD \<longrightarrow> \<not> local.formula (transition.formula (lookup (fst v)))]"
   by auto
  thus ?thesis by(auto simp:del_transitions_impl_def diff_by_label_def)
qed

lemma eliminate_edges_is_disj_invariant_update_P_only[simp]:
  fixes TD :: "'tr list"
    and invariant_condition :: "(('f,'v,'t,'l sharp) disjInv \<Rightarrow> 'tr \<times> ('f,'v,'t,'l sharp) transition \<Rightarrow> shows check)"
  assumes "condition_checker invariant_condition str_c bnd_c rf S NS bnd"
  shows "disj_invariant_update_P_only (eliminate_edges i invariant_condition str_c bnd_c TD)"
unfolding disj_invariant_update_P_only_def
proof(clarify)
  fix P::"('f,'v,'t,'l sharp,'tr) lts_impl"
  fix I::"(('f,'v,'t,'l sharp) disjInv)"
  fix lookup :: "('tr \<Rightarrow> ('f,'v,'t,'l sharp) transition)"
  assume presv:"preserved_lookup lookup P"
  hence lts_impl : "lts_impl P" unfolding preserved_lookup_def by auto
  assume presi:"preserved_invariant I P"
  show "case eliminate_edges i invariant_condition str_c bnd_c TD I lookup P of Inl x \<Rightarrow> True
                     | Inr P2 \<Rightarrow> (cooperation_SN (lts_of P2) \<longrightarrow> cooperation_SN (lts_of P)) \<and>
                                 preserved_lookup lookup P2 \<and> preserved_invariant I P2"
  proof(cases "eliminate_edges i invariant_condition str_c bnd_c TD I lookup P")
    case (Inr P2)
      let ?TD_filt = "filter (\<lambda> v. is_sharp_transition (lookup v) \<and> formula (transition.formula (lookup v))) TD"
      let ?P2 = "del_transitions_impl P ?TD_filt"
      let ?TD_sharp = "set ?TD_filt"
      let ?TD = "map (\<lambda> v. (v,lookup v)) TD"
      have TD: "lookup ` set TD = snd ` set ?TD" by auto
      have P2: "P2 = ?P2" using Inr[unfolded eliminate_edges_def] filter_formula[OF presv] by auto
      have [simp]:"\<And> s lookup. snd ` (\<lambda>v. (v, lookup v)) ` s = lookup ` s" by fastforce
      have okDel:"\<And> \<tau>. \<tau> \<in> set TD \<Longrightarrow> isOK(str_c I (\<tau>, lookup \<tau>)) \<and> isOK(bnd_c I (\<tau>, lookup \<tau>))"
        using Inr[unfolded eliminate_edges_def] by auto
      {
        assume SN: "cooperation_SN (lts_of ?P2)"
        have "cooperation_SN (lts_of P)"
        proof(rule ranking_function_via_disj_invariant)
          show "cooperation_SN (delete_transitions (lts_of P) (lookup ` ?TD_sharp))"
          proof(rule del_transitions_impl)
            show "(\<forall>tr b. tr \<in> fst ` set (transitions_impl P) \<longrightarrow>
                  (tr, b) \<in> set (transitions_impl P) \<longrightarrow> b = lookup tr)"
              using presv[unfolded preserved_lookup_def] by auto next
            from SN show "cooperation_SN (lts_of ?P2)"
            unfolding P2[symmetric].
          qed
          note * = presi[unfolded preserved_invariant_def]
          from * show disj:"disj_invariant (lts_of P) I" by auto
          from * have formOk:"(\<forall>l. list_all formula (I l))" by auto
          show "\<forall>\<tau> \<in> lookup ` ?TD_sharp. is_sharp_transition \<tau>" unfolding TD by auto
          note ** = condition_checkerD[OF assms]
          from ** SN_order_pair_imp_non_inf_order show ninv:"non_inf_order S NS" by auto
          show "\<forall>\<tau> \<in> sharp_transitions_of (lts_of P). \<forall>(s,t) \<in> transition_step \<tau>. \<forall>\<phi> \<in> set (I (location s)).
               (valuation s \<Turnstile> \<phi>) \<longrightarrow> (rf s, rf t) \<in> NS \<union> S"
          proof (intro ballI, unfold split_paired_all prod.case, intro ballI impI)
            fix \<tau> s t \<phi>
            assume \<tau>: "\<tau> \<in> sharp_transitions_of (lts_of P)"
               and st: "(s,t) \<in> transition_step \<tau>"
               and \<phi>: "\<phi> \<in> set (I (location s))"
               and sat: "valuation s \<Turnstile> \<phi>"
            from \<tau> have "\<tau> \<in> transitions (lts_of P)" by auto
            then obtain tr where tr: "(tr,\<tau>) \<in> set (transitions_impl P)" by auto
            show "(rf s, rf t) \<in> NS \<union> S"
            proof (rule **)
              from \<tau> show "is_sharp_transition \<tau>" by auto
              with tr Inr[unfolded eliminate_edges_def]
              show "isOK(invariant_condition I (tr,\<tau>))" by auto
              from tr lts_impl[unfolded lts_impl_def lts_def]
              show "formula (transition.formula \<tau>)" by auto
            qed (insert formOk st \<phi> sat)
          qed
          show "\<forall>\<tau> \<in> lookup ` ?TD_sharp. \<forall>(s,t) \<in> transition_step \<tau>. \<forall>\<phi>\<in>set (I (location s)).
               (valuation s \<Turnstile> \<phi>) \<longrightarrow> (rf s, rf t) \<in> S \<and> (rf s, bnd) \<in> NS \<union> S"
          proof(intro ballI, unfold split_paired_all prod.case, intro ballI impI)
            fix \<tau> s t \<phi>
            assume \<tau>: "\<tau> \<in> lookup ` ?TD_sharp"
               and st: "(s,t) \<in> transition_step \<tau>"
               and \<phi>: "\<phi> \<in> set (I (location s))"
               and sat: "valuation s \<Turnstile> \<phi>"
            from \<tau> obtain tr where tr: "(tr,\<tau>) \<in> set ?TD" by auto
            from tr Inr[unfolded eliminate_edges_def] \<tau>
            have strict:"(rf s, rf t) \<in> S" using **(3) formOk st \<phi> sat by auto
            have bound:"(rf s, bnd) \<in> S"
            proof(rule **(4))
              from \<tau> show "is_sharp_transition \<tau>" "formula (transition.formula \<tau>)" by auto
              from tr Inr[unfolded eliminate_edges_def] show "isOK (bnd_c I (tr, \<tau>))" by auto
            qed (insert formOk st \<phi> sat)
            show "(rf s, rf t) \<in> S \<and> (rf s, bnd) \<in> NS \<union> S" using bound strict by auto
          qed
        qed
      }
      moreover have "preserved_lookup lookup ?P2"
        using preserved_is_preserved_over_del_transitions[OF presv] by auto
      moreover have "preserved_invariant I ?P2"
        using preserved_invariant_is_preserved_over_del_transitions[OF presi] by auto
      ultimately show ?thesis unfolding Inr P2 by auto
  qed auto
qed

(* condition checker of pairs: *)

definition
 "bound_on b s = {(x,y). (x,b) \<in> s \<and> (x,y) \<in> s}"

definition combine_checkers
where
"combine_checkers c1 c2 \<equiv>
  let inv = (\<lambda> i t. do {inv_checker c1 i t
                       ;inv_checker c2 i t});
      strict = (\<lambda> i t. or_ok (str_checker c1 i t) (do {inv_checker c1 i t;str_checker c2 i t})); (* (\<lambda> i t. or_ok (str_c1 i t) (do {inv_c1 i t;str_c2 i t})) *)
      boundc  = (\<lambda> i t. do {bnd_checker c1 i t;bnd_checker c2 i t});
      rf = (\<lambda> s. (ranking_f c1 s,ranking_f c2 s));
      b_s2 = (bound_on (bound c2) (strict_o c2));
      s = lex_two (strict_o c1) (nonstrict_o c1) b_s2;
      ns = lex_two (strict_o c1) (nonstrict_o c1) (nonstrict_o c2 \<union> strict_o c2 \<union> b_s2); (* lex_two S1 NS1 (NS2 \<union> S2 \<union> bound_on bnd2 S2) *)
      bnd = (bound c1,bound c2)
  in (Condition_Checker inv strict boundc rf s ns bnd)"

lemma bound_on_SN:
  assumes "non_inf_order s ns"
  shows "SN (bound_on bnd s)"
proof -
  have sn:"SN {(a, b). (b, bnd) \<in> s \<and> (a, b) \<in> s}" and c1:"s O ns \<subseteq> s" and ts:"trans s"
    using assms non_inf_imp_SN_bound
    unfolding non_inf_order_def by auto
  show ?thesis proof(rule SN_onI) fix f
    assume "\<forall>i. (f i, f (Suc i)) \<in> bound_on bnd s"
    hence all:"\<And>i. (f i, f (Suc i)) \<in> {(x,y). (y,bnd) \<in> ns \<union> s \<and> (x,y) \<in> s}"
      unfolding bound_on_def by auto
    hence s:"\<And> i. (f i, f (Suc i)) \<in> s"
      and b:"\<And> i. (f (Suc i),bnd) \<in> ns \<union> s" by auto
    from sn[unfolded SN_defs]
      obtain i where i:"(f i, f (Suc i)) \<notin> {(a, b). (b, bnd) \<in> s \<and> (a, b) \<in> s}" by auto
    hence "(f (Suc i), bnd) \<notin> s" using s by auto
    hence "(f (Suc i), f (Suc (Suc i))) \<notin> {(x,y). (y,bnd) \<in> ns \<union> s \<and> (x,y) \<in> s}"
      using c1 ts[unfolded trans_def,rule_format] by auto
    thus False using all by auto
  qed
qed 

lemma bound_on_order_pair:
  assumes "order_pair s ns"
  shows "order_pair (bound_on bnd s) ns"
proof
  from assms[unfolded order_pair_def pre_order_pair_def compat_pair_def]
  have *:"refl ns" "trans s" "trans ns" "ns O s \<subseteq> s" "s O ns \<subseteq> s" by auto
  thus "refl ns" "trans ns" "ns O bound_on bnd s \<subseteq> bound_on bnd s"
    "bound_on bnd s O ns \<subseteq> bound_on bnd s"
    unfolding bound_on_def by auto
  show "trans (bound_on bnd s)" proof(rule transI) fix x y z
    assume "(x, y) \<in> bound_on bnd s" "(y, z) \<in> bound_on bnd s"
    hence "(x,y) \<in> s" "(y,z) \<in> s" "(x, bnd) \<in> s" unfolding bound_on_def by auto
    from *(2)[unfolded trans_def,rule_format,OF this(1,2)] this(3)
    show "(x, z) \<in> bound_on bnd s" unfolding bound_on_def by auto qed
qed

lemma lex_two_non_inf:
  assumes "SN_order_pair S2 NS2"
      and "non_inf_order S1 NS1"
  shows "non_inf_order (lex_two S1 NS1 S2) (lex_two S1 NS1 NS2)"
proof
  have "refl NS2" "trans NS2" "NS2 O S2 \<subseteq> S2" "S2 O NS2 \<subseteq> S2" "trans NS1" "refl NS1" "trans S2" and
       sn:"SN S2"         and
       m1:"S1 O NS1 \<subseteq> S1" and
       m2:"NS1 O S1 \<subseteq> S1" and
       tr1:"trans S1" and
       ni:"non_inf S1"    
    using assms
    unfolding non_inf_order_def SN_order_pair_def order_pair_def
              compat_pair_def pre_order_pair_def SN_ars_def
    by auto
  thus *:"lex_two S1 NS1 S2 O lex_two S1 NS1 NS2 \<subseteq> lex_two S1 NS1 S2"
       "lex_two S1 NS1 NS2 O lex_two S1 NS1 S2 \<subseteq> lex_two S1 NS1 S2"
       "trans (lex_two S1 NS1 NS2)" "refl (lex_two S1 NS1 NS2)" "trans (lex_two S1 NS1 S2)"
    using lex_two_order_pair[unfolded order_pair_def compat_pair_def pre_order_pair_def]
    by metis+
  from tr1[unfolded trans_def,rule_format] m1
  have tr1a:"\<And> x y z. (x, y) \<in> S1 \<Longrightarrow> (y, z) \<in> NS1 \<union> S1 \<Longrightarrow> (x, z) \<in> S1" by blast
  have sn:"\<And> f. \<exists>i. (f i, f (Suc i)) \<notin> S2" using sn[unfolded SN_on_def] by auto
  show "non_inf (lex_two S1 NS1 S2)" proof(rule non_infI) fix f fix a::"'b\<times>'a"
    assume "\<And>i. (f i, f (Suc i)) \<in> lex_two S1 NS1 S2"
    hence "\<And>i. (\<lambda>(a1, b1) (a2, b2). (a1, a2) \<in> S1 \<or> (a1, a2) \<in> NS1 \<and> (b1, b2) \<in> S2) (f i) (f (Suc i))"
       unfolding lex_two.simps by auto
    hence i:"\<And>i. (fst (f i), fst (f (Suc i))) \<in> S1 \<or> (fst (f i), fst (f (Suc i))) \<in> NS1 \<and> (snd (f i), snd (f (Suc i))) \<in> S2"
       by (simp add: case_prod_beta)
    hence "\<And>i. (fst (f i), fst (f (Suc i))) \<in> NS1 \<union> S1"
       by auto
    from non_inf_order.chain_split[OF assms(2) basic_monos(1) basic_monos(1),folded not_INFM] this
    have "\<not> (\<exists>\<^sub>\<infinity>i. (fst (f i), fst (f (Suc i))) \<in> S1) \<or> \<not> (\<exists>\<^sub>\<infinity>i. (fst (f i), fst a) \<in> NS1 \<union> S1)"
      by auto
    then consider (a) "\<not> (\<exists>\<^sub>\<infinity>i. (fst (f i), fst a) \<in> NS1 \<union> S1)"
                | (b) "\<not> (\<exists>\<^sub>\<infinity>i. (fst (f i), fst (f (Suc i))) \<in> S1)" by blast
    thus "\<exists>i. (f i, a) \<notin> lex_two S1 NS1 S2" proof(cases)
    case a
      from this[unfolded not_INFM MOST_nat_le]
      obtain i where "\<And> j. j \<ge> i \<Longrightarrow> (fst (f j), fst a) \<notin> S1" by auto
      hence "\<And> j. (fst (f (j+i)), fst a) \<notin> S1" by auto
      hence "\<exists> j. ((fst (f (j+i)),snd (f (j+i))), (fst a,snd a)) \<notin> lex_two S1 NS1 S2"
        using sn[of "\<lambda> j. snd (f (j + i))"] i
        by (auto simp:fst_def snd_def) (metis Un_iff add_Suc tr1a)
      hence "\<exists> j. ((fst (f j),snd (f j)), (fst a,snd a)) \<notin> lex_two S1 NS1 S2" by blast
      thus ?thesis by auto next
    case b
      hence "\<exists> i. \<forall> j. j \<ge> i \<longrightarrow> (fst (f j), fst (f (Suc j))) \<notin> S1" using sn INFM_nat_le by simp
      then obtain i where "\<And> j. j \<ge> i \<Longrightarrow> (snd (f j), snd (f (Suc j))) \<in> S2" using i by auto
      from this[of "_ + i",simplified]
      have "\<And> j. (snd (shift f i j), snd (shift f i (Suc j))) \<in> S2" by simp
      hence "\<exists>f. f 0 \<in> UNIV \<and> (\<forall>i. (f i, f (Suc i)) \<in> S2)" by fast
      hence False using sn by auto
      thus ?thesis ..
    qed
  qed
qed

lemma SN_order_pair_union: assumes "SN_order_pair S NS" shows "SN_order_pair S (NS \<union> S)"
proof
  note * = assms[unfolded SN_order_pair_def order_pair_def SN_ars_def pre_order_pair_def compat_pair_def]
  show "trans S" "SN S" "refl (NS \<union> S)"
    using * by (auto simp: refl_O_iff sup.coboundedI1)
  show comp:"(NS \<union> S) O S \<subseteq> S" "S O (NS \<union> S) \<subseteq> S"
    using * by (auto simp:trans_def)
  show "trans (NS \<union> S)" using comp * unfolding trans_def by blast
qed
lemma non_inf_union: assumes "non_inf_order S NS" shows "non_inf_order S (NS \<union> S)"
proof
  note * = assms[unfolded non_inf_order_def]
  show "trans S" "non_inf S" "refl (NS \<union> S)"
    using * by (auto simp: refl_O_iff sup.coboundedI1)
  show comp:"(NS \<union> S) O S \<subseteq> S" "S O (NS \<union> S) \<subseteq> S" 
    using * trans_def[of S] by blast+
  show "trans (NS \<union> S)" using comp * unfolding trans_def by blast
qed

lemma condition_checker_combined:
 assumes c1:"condition_checker inv_c1 str_c1 bnd_c1 rf1 S1 NS1 bnd1"
     and c2:"condition_checker inv_c2 str_c2 bnd_c2 rf2 S2 NS2 bnd2"
     and c3:"(Condition_Checker inv_c3 str_c3 bnd_c3 rf3 S3 NS3 bnd3)
            = combine_checkers (Condition_Checker inv_c1 str_c1 bnd_c1 rf1 S1 NS1 bnd1)
                               (Condition_Checker inv_c2 str_c2 bnd_c2 rf2 S2 NS2 bnd2)"
 shows "condition_checker inv_c3 str_c3 bnd_c3 rf3 S3 NS3 bnd3"
proof
  have d3:"inv_c3 = (\<lambda> i t. do {inv_c1 i t
                               ;inv_c2 i t})"
          "str_c3 = (\<lambda> i t. or_ok (str_c1 i t) (do {inv_c1 i t;str_c2 i t}))"
          "bnd_c3 = (\<lambda> i t. do {bnd_c1 i t;bnd_c2 i t})"
          "rf3 = (\<lambda>s. (rf1 s, rf2 s))"
          "S3 = lex_two S1 NS1 (bound_on bnd2 S2)"
          "NS3 = lex_two S1 NS1 (NS2 \<union> S2 \<union> bound_on bnd2 S2)"
          "bnd3 = (bnd1,bnd2)"
    using c3[unfolded combine_checkers_def Let_def] by auto
  have ni:"non_inf_order S1 NS1" "non_inf_order S2 NS2"
    using c1 c2 unfolding condition_checker_def by auto
  hence "SN (bound_on bnd2 S2)" using bound_on_SN by force
  hence "SN_order_pair (bound_on bnd2 S2) (NS2 \<union> S2)"
    using non_inf_union[OF ni(2),unfolded non_inf_order_def] bound_on_order_pair
    unfolding order_pair_def SN_ars_def pre_order_pair_def compat_pair_def SN_order_pair_def
    by metis
  from SN_order_pair_union[OF this]
  show "non_inf_order S3 NS3" unfolding d3 using ni(1) lex_two_non_inf by blast
  fix tr \<tau>
  fix I:: "'a sharp \<Rightarrow> ('f, 'b \<times> 't) Term.term list list list"
  fix s t \<phi>
  assume a:"\<forall>l. list_all local.formula (I l)"
           "local.formula (transition.formula \<tau>)"
           "is_sharp_transition \<tau> "
           "(s, t) \<in> transition_step \<tau> "
           " \<phi> \<in> set (I (location s)) "
           " valuation s \<Turnstile> \<phi>"
  { assume "isOK (inv_c3 I (tr, \<tau>))"
    hence ok:"isOK (inv_c1 I (tr, \<tau>))" "isOK (inv_c2 I (tr, \<tau>))" unfolding d3 by auto
    have b1:"(rf1 s, rf1 t) \<in> NS1 \<union> S1" using condition_checkerD(2)[OF c1 a(1-3) ok(1) a(4-)].
    have b2:"(rf2 s, rf2 t) \<in> NS2 \<union> S2" using condition_checkerD(2)[OF c2 a(1-3) ok(2) a(4-)].
    show "(rf3 s, rf3 t) \<in> NS3 \<union> S3" using b1 b2 unfolding d3 bound_on_def by auto
  }
  { assume "isOK (bnd_c3 I (tr,\<tau>))"
    hence ok:"isOK (bnd_c1 I (tr, \<tau>))" "isOK (bnd_c2 I (tr, \<tau>))" unfolding d3 by auto
    have b1:"(rf1 s, bnd1) \<in> S1" using condition_checkerD(4)[OF c1 a(1-3) ok(1) a(4-)].
    have b2:"(rf2 s, bnd2) \<in> S2" using condition_checkerD(4)[OF c2 a(1-3) ok(2) a(4-)].
    show "(rf3 s, bnd3) \<in> S3" using b1 b2 unfolding d3 bound_on_def by auto
    assume "isOK (str_c3 I (tr, \<tau>))"
    hence "isOK (str_c1 I (tr, \<tau>)) \<or> (isOK (inv_c1 I (tr, \<tau>)) \<and> isOK (str_c2 I (tr, \<tau>)))"
      using d3 or_is_or by blast
    hence r2:"(rf1 s, rf1 t) \<in> S1 \<or> ((rf1 s, rf1 t) \<in> NS1 \<union> S1 \<and> (rf2 s, rf2 t) \<in> S2)"
      using condition_checkerD(3)[OF c1 a(1-3) _ _ a(4-)]
            condition_checkerD(2)[OF c1 a(1-3) _ a(4-)]
            condition_checkerD(3)[OF c2 a(1-3) _ _ a(4-)] ok by auto
    thus "(rf3 s, rf3 t) \<in> S3" using b2 unfolding d3 bound_on_def by auto
  }
qed

end


end