theory Formula
imports
  "$AFP/Abstract-Rewriting/Abstract_Rewriting"
  Term
  Trs_Impl
  Option_Monad
begin

type_synonym ('f,'v,'t) exp = "('f, 'v \<times> 't) term"
type_synonym ('v,'t,'d) valuation = "'v \<times> 't \<Rightarrow> 'd"

type_synonym ('f,'v,'t) formula = "('f,'v,'t) exp list list"

abbreviation disj_atoms :: "('f,'v,'t) exp list \<Rightarrow> ('f,'v,'t) formula" where
  "disj_atoms \<equiv> (\<lambda> as. [as])"
abbreviation conj_atoms :: "('f,'v,'t) exp list \<Rightarrow> ('f,'v,'t) formula" where
  "conj_atoms \<equiv> (\<lambda> as. (map (\<lambda> a. [a]) as))"

abbreviation form_False :: "('f,'v,'t) formula" where "form_False \<equiv> [[]]"
abbreviation form_True  :: "('f,'v,'t) formula" where "form_True \<equiv> []"

definition form_and :: "('f,'v,'t) formula \<Rightarrow> ('f,'v,'t) formula \<Rightarrow> ('f,'v,'t) formula"
  (infix "\<and>\<^sub>f" 80) where "\<phi> \<and>\<^sub>f \<psi> \<equiv> \<phi> @ \<psi>"

definition "vars_formula \<phi> \<equiv> \<Union>cl \<in> set \<phi>. \<Union>a \<in> set cl. vars_term a"

consts rename_vars :: "('a \<Rightarrow> 'b) \<Rightarrow> 'c \<Rightarrow> 'd"

definition rename_vars_exp :: "('v \<Rightarrow> 'w) \<Rightarrow> ('f,'v,'t) exp \<Rightarrow> ('f,'w,'t) exp" where
  "rename_vars_exp r = map_term id (map_prod r id)"

adhoc_overloading rename_vars rename_vars_exp

lemma rename_vars_Fun: "rename_vars r (Fun f ts) = Fun f (map (rename_vars r) ts)"
  unfolding rename_vars_exp_def by auto

definition rename_vars_formula :: "('v \<Rightarrow> 'w) \<Rightarrow> ('f,'v,'t) formula \<Rightarrow> ('f,'w,'t) formula" where
  "rename_vars_formula r cls \<equiv> map (map (rename_vars r)) cls"

adhoc_overloading rename_vars rename_vars_formula

record ('f,'t,'d) "theory" =
  types :: "'f \<Rightarrow> 't list \<times> 't"
  type_values :: "'t \<Rightarrow> 'd set"
  bool :: "'t"
  to_bool :: "'d \<Rightarrow> bool"
  "interpret" :: "'f \<Rightarrow> 'd list \<Rightarrow> 'd"

locale "pretheory" =
  fixes T :: "('f,'t,'d) theory"
begin

  abbreviation types where "types \<equiv> theory.types T"
  abbreviation type_values where "type_values \<equiv> theory.type_values T"
  abbreviation bool where "bool \<equiv> theory.bool T"
  abbreviation to_bool where "to_bool \<equiv> theory.to_bool T"
  abbreviation "interpret" where "interpret \<equiv> theory.interpret T"

  definition "assignment \<alpha> \<equiv> \<forall>x t. \<alpha> (x,t) \<in> type_values t"

  lemma assignmentI[intro]:
    assumes "\<And>x t. \<alpha> (x,t) \<in> type_values t" shows "assignment \<alpha>"
    using assms by (auto simp: assignment_def)

  lemma assignment_range[dest]: "assignment \<alpha> \<Longrightarrow> \<alpha> (x,t) \<in> type_values t"
    by (auto simp: assignment_def)

  fun type_of :: "('f,'v,'t) exp \<Rightarrow> 't option" where
    "type_of (Var v) = Some (snd v)" |
    "type_of (Fun f ts) = do {
        tys \<leftarrow> mapM type_of ts;
        let (tys',t) = types f;
        guard (tys = tys');
        Some t
      }"

  definition "exp e \<equiv> type_of e \<noteq> None"
  abbreviation "EXP \<equiv> Collect exp"

  definition "is_bool e \<equiv> type_of e = Some bool \<and> \<not> is_Var e"
  abbreviation "BOOL_EXP \<equiv> Collect is_bool"

  lemma is_bool_imp_exp[simp]: "is_bool e \<Longrightarrow> exp e" unfolding is_bool_def exp_def by auto

  lemma type_of_fun: assumes "type_of (Fun f ts) \<noteq> None"
    shows "type_of (Fun f ts) = Some (snd (types f))"
    and "mapM type_of ts = Some (fst (types f))"
  proof -
    obtain v where a:"type_of (Fun f ts) = Some v" using assms by auto
    have "(let (tys', t) = types f in
          guard (the (Option_Monad.mapM type_of ts) = tys') \<bind> (\<lambda>_. Some t)) =
          Some v" using a unfolding type_of.simps by auto
    hence "guard (the (Option_Monad.mapM type_of ts) = fst (types f))
          \<bind> (\<lambda>_. Some (snd (types f))) =
          Some v" by (simp add: prod.case_eq_if)
    thus "type_of (Fun f ts) = Some (snd (types f))"
     and "mapM type_of ts = Some (fst (types f))" using a by auto
  qed

  lemma type_of_rename_vars[simp]: "type_of (rename_vars r a) = type_of a"
  proof (induct a)
    case (Fun f ts)
    obtain tys' t where ty: "types f = (tys',t)" by force
    have id: "Option_Monad.mapM type_of (map (rename_vars r) ts) = Option_Monad.mapM type_of ts"
      using Fun by (induct ts, auto)
    show ?case using ty by (auto simp: rename_vars_Fun id)
  qed (simp add: rename_vars_exp_def)

  lemma is_bool_rename_vars[simp]: "is_bool (rename_vars r a) = is_bool a"
    unfolding is_bool_def type_of_rename_vars
    by(cases a,auto simp:rename_vars_exp_def)

  fun typelist_match where
    "typelist_match (Cons a as) (Cons ty tys) = (a \<in> type_values ty \<and> typelist_match as tys)" |
    "typelist_match [] [] = True" |
    "typelist_match _ _ = False"

  definition formula :: "('f,'v,'t) formula \<Rightarrow> bool" where
    "formula \<phi> \<equiv> \<forall>cl \<in> set \<phi>. \<forall>lit \<in> set cl. is_bool lit"

  lemma formulaI[intro]:
    assumes "\<And>cl lit. cl \<in> set \<phi> \<Longrightarrow> lit \<in> set cl \<Longrightarrow> is_bool lit"
    shows "formula \<phi>" using assms by (auto simp: formula_def)

  lemma formulaD[dest]:
    assumes "formula \<phi>"
    shows "cl \<in> set \<phi> \<Longrightarrow> lit \<in> set cl \<Longrightarrow> is_bool lit" using assms by (auto simp: formula_def)

  lemma formulaE[elim]:
    assumes "formula \<phi>" "(\<And>cl lit. cl \<in> set \<phi> \<Longrightarrow> lit \<in> set cl \<Longrightarrow> is_bool lit) \<Longrightarrow> P"
    shows "P"
    using assms by (auto simp: formula_def)


  lemma formula_rename_vars[simp]: "formula (rename_vars r f) = formula f"
    by (auto simp: formula_def rename_vars_formula_def)
  lemma formula_rename_vars_map[simp]:
    "list_all formula (map (rename_vars r) l) =  list_all formula l"
    using formula_rename_vars unfolding pred_list_def by auto
  lemma formula_and[simp]: "formula (\<phi> \<and>\<^sub>f \<psi>) \<longleftrightarrow> formula \<phi> \<and> formula \<psi>" by (auto simp: formula_def form_and_def)
  lemma formula_andI[intro]: "formula \<phi> \<Longrightarrow> formula \<psi> \<Longrightarrow> formula (\<phi> \<and>\<^sub>f \<psi>)" by auto
  lemma formula_False[simp]: "formula form_False" by (auto simp: formula_def)
  lemma formula_True[simp]: "formula form_True" by (auto simp: formula_def)

  context
    fixes \<alpha> :: "('v,'t,'d) valuation"
  begin

  fun eval :: "('f,'v,'t) exp \<Rightarrow> 'd" where
    "eval (Var v) = \<alpha> v" |
    "eval (Fun f ts) = interpret f (map eval ts)"

  definition "eval_true = to_bool \<circ> eval"

  definition satisfies :: "('f,'v,'t) formula \<Rightarrow> bool"
    where "satisfies \<phi> \<equiv> (\<forall>cl \<in> set \<phi>. \<exists>lit \<in> set cl. eval_true lit)"

  lemma satisfiesI[intro]:
    assumes "\<And>cl. cl \<in> set \<phi> \<Longrightarrow> \<exists>lit \<in> set cl. eval_true lit"
    shows "satisfies \<phi>"
    using assms unfolding satisfies_def by auto

  lemma satisfiesD[dest]:
    assumes "satisfies \<phi>"
    shows "\<And>cl. cl \<in> set \<phi> \<Longrightarrow> \<exists>lit \<in> set cl. eval_true lit"
    using assms unfolding satisfies_def by auto

  lemma satisfiesE[elim]:
    assumes "satisfies \<phi>"
        and "(\<And>cl. cl \<in> set \<phi> \<Longrightarrow> \<exists>lit \<in> set cl. eval_true lit) \<Longrightarrow> P"
    shows P
    using assms unfolding satisfies_def by auto

  lemma form_True[simp]: "satisfies form_True" by (auto simp: satisfies_def)

  lemma satisfies_andI[intro]:
    assumes "satisfies \<phi>" and "satisfies \<psi>" shows "satisfies (\<phi> \<and>\<^sub>f \<psi>)"
    using assms by (auto simp: form_and_def satisfies_def)

  lemma satisfies_andD[dest]:
    assumes "satisfies (\<phi> \<and>\<^sub>f \<psi>)"  shows "satisfies \<phi>" and "satisfies \<psi>"
    using assms by (auto simp: form_and_def satisfies_def)

  end

  notation satisfies (infix "\<Turnstile>" 40)

  lemma eval_with_fresh_var[simp]: "xx \<notin> vars_term e \<Longrightarrow> eval (\<alpha>(xx:=a)) e = eval \<alpha> e"
  proof(induct e)
    case (Var x) then show ?case by auto
  next
    case (Fun f es)
      show ?case unfolding eval.simps
      proof(rule cong[of "interpret f", OF refl], unfold map_eq_conv, intro ballI)
        fix e' assume "e' \<in> set es"
        moreover with Fun(2) have "xx \<notin> vars_term e'" by auto
        ultimately show "eval (\<alpha>(xx:=a)) e' = eval \<alpha> e'" by (intro Fun(1), auto)
      qed
  qed

  lemma satisfies_with_fresh_var[simp]:
    assumes fv: "xx \<notin> vars_formula \<phi>"
    shows "\<alpha>(xx:=a) \<Turnstile> \<phi> \<longleftrightarrow> \<alpha> \<Turnstile> \<phi>" (is "?l \<longleftrightarrow> ?r")
  proof(intro iffI satisfiesI)
    fix cl assume cl: "cl \<in> set \<phi>"
    { assume ?r
      with cl obtain lit where lit: "lit \<in> set cl" and eval: "eval_true \<alpha> lit" by auto
      from fv cl lit have "xx \<notin> vars_term lit" by (auto simp: vars_formula_def)
      with eval have "eval_true (\<alpha>(xx:=a)) lit" by (simp add: eval_true_def)
      with lit show "\<exists>lit \<in> set cl. eval_true (\<alpha>(xx:=a)) lit" by auto
    }
    { assume ?l
      with cl obtain lit where lit: "lit \<in> set cl" and eval: "eval_true (\<alpha>(xx:=a)) lit" by auto
      from fv cl lit have "xx \<notin> vars_term lit" by (auto simp: vars_formula_def)
      with eval have "eval_true \<alpha> lit" by (simp add: eval_true_def)
      with lit show "\<exists>lit \<in> set cl. eval_true \<alpha> lit" by auto
    }
  qed

  context
    fixes \<alpha> :: "('v,'t,'d) valuation" and \<beta> :: "('w,'t,'d) valuation" and r :: "'v \<Rightarrow> 'w"
    assumes r: "\<And> x t. \<alpha> (x,t) = \<beta> (r x, t)"
  begin

  lemma eval_rename_vars: "eval \<beta> (rename_vars r a) = eval \<alpha> a"
  proof (induct a)
    case (Var x)
    show ?case by (cases x, auto simp: rename_vars_exp_def r)
  next
    case (Fun f ts)
    thus ?case by (auto simp: rename_vars_Fun intro: arg_cong[of _ _ "interpret f"])
  qed

  lemma eval_true_rename_vars[simp]: "eval_true \<beta> (rename_vars r \<phi>) = eval_true \<alpha> \<phi>"
    by (auto simp: eval_true_def eval_rename_vars)

  lemma assignment_rename_vars: "assignment \<beta> \<Longrightarrow> assignment \<alpha>"
    unfolding assignment_def using r by auto

  lemma satisfies_rename_vars: "\<beta> \<Turnstile> rename_vars r \<phi> \<Longrightarrow> \<alpha> \<Turnstile> \<phi>"
    using formula_rename_vars by (auto simp: satisfies_def assignment_rename_vars rename_vars_formula_def)
  end

  definition implies :: "('f,'v,'t) formula \<Rightarrow> ('f,'v,'t) formula \<Rightarrow> bool" (infix "\<longrightarrow>\<^sub>f" 60) where
    "\<phi> \<longrightarrow>\<^sub>f \<psi> \<equiv> \<forall> \<alpha>. formula \<phi> \<longrightarrow> formula \<psi> \<longrightarrow> assignment \<alpha> \<longrightarrow> \<alpha> \<Turnstile> \<phi> \<longrightarrow> \<alpha> \<Turnstile> \<psi>"

  lemma impliesI[intro]:
    assumes "\<And> \<alpha>. formula \<phi> \<Longrightarrow> formula \<psi> \<Longrightarrow> assignment \<alpha> \<Longrightarrow> \<alpha> \<Turnstile> \<phi> \<Longrightarrow> \<alpha> \<Turnstile> \<psi>"
    shows "\<phi> \<longrightarrow>\<^sub>f \<psi>" using assms by (auto simp: implies_def)

  lemma impliesD[dest]:
    assumes "\<phi> \<longrightarrow>\<^sub>f \<psi>"
    shows "formula \<phi> \<Longrightarrow> formula \<psi> \<Longrightarrow> assignment \<alpha> \<Longrightarrow> \<alpha> \<Turnstile> \<phi> \<Longrightarrow> \<alpha> \<Turnstile> \<psi>"
    using assms by (auto simp: implies_def)

  lemma impliesE[elim]:
    assumes "\<phi> \<longrightarrow>\<^sub>f \<psi>" "(\<And>\<alpha>. formula \<phi> \<Longrightarrow> formula \<psi> \<Longrightarrow> assignment \<alpha> \<Longrightarrow> \<alpha> \<Turnstile> \<phi> \<Longrightarrow> \<alpha> \<Turnstile> \<psi>) \<Longrightarrow> P"
    shows P using assms by (auto simp: implies_def)


end

locale "theory" = "pretheory" T for T :: "('f,'t,'d) theory" +
  assumes well_typed:
    "typelist_match ds (fst (types f)) \<Longrightarrow> interpret f ds \<in> type_values (snd (types f))"
  and consistent: "type_values ty \<noteq> {}"
begin

  lemma eval_types:
    assumes "type_of x = Some v"
    and "assignment \<alpha>"
    shows "eval \<alpha> x \<in> type_values v"
  using assms proof (induct x arbitrary: v)
    case (Fun f ls)
      obtain tps where tps: "fst (types f) = tps" by auto
      with Fun.prems have "mapM type_of ls = Some tps" by (auto intro: type_of_fun(2))
      hence tlm: "typelist_match (map (eval \<alpha>) ls) (fst (types f))" unfolding tps
        using Fun.hyps assms(2)
      proof(induct "(map (eval \<alpha>) ls)" tps arbitrary: ls f rule: typelist_match.induct)
        case (1 a as b bs)
          obtain l ls' where ls:"ls = Cons l ls'" using "1.hyps"(2) by auto
          have typ_l_b : "type_of l = Some b" using "1.prems"(1) "1.hyps"(2) 
            unfolding ls "1.hyps"(2)[symmetric] by simp
          have as:"as = map (eval \<alpha>) ls'" using "1.hyps"(2) ls by auto
          have map:"Option_Monad.mapM type_of ls' = Some bs" using "1.prems" ls by auto
          have "a = (eval \<alpha> l)" using "1.hyps"(2) ls by auto
          hence "eval \<alpha> l \<in> type_values b" using "1.prems"(2) typ_l_b assms(2) ls by auto
          thus ?case unfolding list.map typelist_match.simps
            using "1.hyps"(1)[of ls',OF as map] "1.prems" list.set_intros(2) ls by auto
      qed(auto)
      show ?case using well_typed[OF tlm] unfolding "eval.simps"
        using Fun.prems(1) option.discI option.sel type_of_fun(1) by fastforce next
    case (Var var)
      have v:"v = snd var" using Var.prems(1) unfolding type_of.simps by auto
      show ?case using assms(2)
        unfolding eval.simps assignment_def v using prod.collapse by metis
  qed
end

type_synonym ('h,'f,'v,'t) theory_checker = 
  "'h \<Rightarrow> ('f,'v,'t) exp list \<Rightarrow> ('f,'v,'t) exp list \<Rightarrow> shows check"

locale pre_theory_checker = pretheory T
  for T :: "('f::show,'t::show,'d) theory"
  and theory_checker :: "('h::show,'f,'v::show,'t) theory_checker"
begin

definition formula_checker ::
    "'h list  \<Rightarrow> ('f,'v,'t) formula \<Rightarrow> ('f,'v,'t) formula \<Rightarrow> shows check"
where
"formula_checker hints \<phi> \<psi> \<equiv>
  if [] \<in> set \<phi> then succeed else do {
     check_allm (\<lambda> cl. check (length cl = 1) (shows ''disjunctions in premises not (yet) supported'')) \<phi>;
     check (length hints = length \<psi>) (shows ''number of hints does not match number of conclusions'');
     let prems = map hd \<phi>;
     check_allm (\<lambda> (h,c). theory_checker h prems c <+? (\<lambda> s. shows ''The conclusion '' o shows c o shows '' does not follow from the premises '' o shows prems o shows '' via the hints '' o shows h o shows ''\<newline>'' o s)) (zip hints \<psi>)
   }"


(* Check_transition will check validity of an expression given:
    - a disjunctive invariant
    - another formula that holds (presumably the transition formula)
*)
fun check_transition ::
  "( ('f,'v,'t) formula \<Rightarrow> shows + 'h list \<times> 'h list)
  \<Rightarrow> ('f,'v,'t) formula list (* precondition *)
  \<Rightarrow> ('f,'v,'t) formula (* transition condition *)
  \<Rightarrow> ('f,'v,'t) exp
  \<Rightarrow> shows check"  where
"check_transition hinter (Cons invar invars) \<phi> res
  = do { (invar_hint,trans_hint) <- hinter invar
       ; check (length invar_hint = length invar) (shows ''Length of invariant-hint and should match number of invariant-conjuncts'')
       ; check (length trans_hint = length \<phi>) (shows ''Length of transition-hint and number of transition-conjuncts should match'')
       ; formula_checker (invar_hint @ trans_hint) (invar @ \<phi>) [[res]]
       ; check_transition hinter invars \<phi> res} <+? (\<lambda> s. shows ''checking a disjunct in the invariant\<newline>'' \<circ> s)" | 
"check_transition hinter Nil \<phi> res = Inr ()"
end

locale theory_checker =
  "theory" T + pre_theory_checker T theory_checker
  for T :: "('f::show,'t::show,'d) theory"
  and theory_checker :: "('h::show,'f,'v::show,'t) theory_checker" +
  assumes theory_checker: "isOK(theory_checker h prems concls)
    \<Longrightarrow> set prems \<subseteq> BOOL_EXP
    \<Longrightarrow> set concls \<subseteq> BOOL_EXP
    \<Longrightarrow> assignment \<alpha>
    \<Longrightarrow> \<forall>prem \<in> set prems. eval_true \<alpha> prem
    \<Longrightarrow> \<exists>concl \<in> set concls. eval_true \<alpha> concl"
begin

lemma theory_checker_implies: assumes "isOK(theory_checker h prems concls)"
   "set prems \<subseteq> BOOL_EXP"
   "set concls \<subseteq> BOOL_EXP"
   shows "conj_atoms prems \<longrightarrow>\<^sub>f disj_atoms concls"
proof
  fix \<alpha> assume \<alpha>: "assignment \<alpha>" and models: "\<alpha> \<Turnstile> conj_atoms prems"
  show "\<alpha> \<Turnstile> disj_atoms concls"
  proof(intro satisfiesI)
    fix cl assume cl: "cl \<in> set (disj_atoms concls)"
    from satisfiesD[OF models] have "\<forall>prem \<in> set prems. eval_true \<alpha> prem"
    apply (auto simp: image_def) by (metis in_set_simps(2))
    from theory_checker[OF assms \<alpha> this] cl show "\<exists>lit\<in>set cl. eval_true \<alpha> lit" by auto
  qed
qed


lemma formula_checker:
  assumes ok: "isOK(formula_checker hints \<phi> \<psi>)"
  shows "\<phi> \<longrightarrow>\<^sub>f \<psi>"
proof
  fix \<alpha>
  assume \<alpha>: "assignment \<alpha>" and \<alpha>\<phi>: "\<alpha> \<Turnstile> \<phi>" and \<phi>: "formula \<phi>" and \<psi>: "formula \<psi>"
  show "\<alpha> \<Turnstile> \<psi>"
  proof
    fix cl
    assume cl: "cl \<in> set \<psi>"
    then obtain i where i: "i < length \<psi>" and cli: "cl = \<psi> ! i"
      unfolding set_conv_nth by auto
    from \<alpha>\<phi> have empty: "[] \<notin> set \<phi>" by auto
    note ok = this ok[unfolded formula_checker_def]
    from ok have "length hints = length \<psi>" by auto
    with i have "(hints ! i, \<psi> ! i) \<in> set (zip hints \<psi>)" unfolding set_zip
      by auto
    then obtain h where "(h, cl) \<in> set (zip hints \<psi>)" unfolding cli by blast
    with ok have tc': "isOK (theory_checker h (map hd \<phi>) cl)" by auto
    show "Bex (set cl) (eval_true \<alpha>)"
    proof (rule theory_checker[OF tc' _ _ \<alpha>])
      show "set cl \<subseteq> BOOL_EXP" using cl \<psi> by (auto simp: formula_def)
      show "set (map hd \<phi>) \<subseteq> BOOL_EXP" using empty \<phi> by (induct \<phi>, auto simp: formula_def)
      from ok have "\<And> x. x \<in> set \<phi> \<Longrightarrow> length x = Suc 0" by auto
      thus "\<forall>prem \<in> set (map hd \<phi>). eval_true \<alpha> prem" using \<alpha>\<phi> unfolding set_map
      proof (induct \<phi>)
        case (Cons c cls)
        from Cons(2)[of c] obtain d where c: "c = [d]" by (cases c, cases "tl c", auto)
        from Cons(3) have "\<alpha> \<Turnstile> cls" by (elim satisfiesE, auto)
        with Cons show ?case by (auto simp:c)
      qed auto
    qed
  qed
qed

lemma check_transition_OK:
  assumes "isOK (check_transition hinter disj_inv \<phi> e)"
  and formula_ok: "formula \<phi>"
  and disj_inv_ok: "list_all formula disj_inv"
  and inset: "\<phi>' \<in> set disj_inv"
  and \<alpha>: "assignment \<alpha>"
  and sats: "\<alpha> \<Turnstile> \<phi>" "\<alpha> \<Turnstile> \<phi>'"
  and typeOk: "is_bool e"
  shows "eval_true \<alpha> e"
using assms(1-4) proof(induct disj_inv)
  case (Cons invar as) thus ?case
    proof(cases "invar = \<phi>'")
      case True
        obtain invar_hint trans_hint where obtl:"Inr (invar_hint,trans_hint) = hinter invar"
          and okC: "isOK(formula_checker (invar_hint @ trans_hint) (\<phi>' \<and>\<^sub>f \<phi>) [[e]])"
          using Cons unfolding True form_and_def by auto
        from Cons True have "formula \<phi>'" by auto
        moreover from sats have sat: "\<alpha> \<Turnstile> \<phi>' \<and>\<^sub>f \<phi>" by auto
        moreover have "\<phi>' \<and>\<^sub>f \<phi> \<longrightarrow>\<^sub>f disj_atoms [e]" by(rule formula_checker[OF okC])
        ultimately have "\<alpha> \<Turnstile> disj_atoms [e]" using \<alpha> formula_ok typeOk by (elim impliesD,auto)
        then show ?thesis by auto
    qed(auto)
qed(auto)

end

fun or_ok :: "'a check \<Rightarrow> 'a check \<Rightarrow> 'a check" where
"or_ok (Inl a) b = b" |
"or_ok (Inr a) b = Inr a" 

lemma or_is_or: "isOK (or_ok a b) = isOK a \<or> isOK b" using or_ok.elims by blast

end